from functools import partial
from itertools import zip_longest
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union, cast

import numpy as np
from more_itertools import batched
from numpy.typing import ArrayLike, NDArray
from tqdm.auto import tqdm

from ._types import Reader
from ._utils import _normalize_contig_name

__all__ = []
try:
    import tensorstore as ts
    import zarr

    ZARR_TENSORSTORE_INSTALLED = True
except ImportError:
    ZARR_TENSORSTORE_INSTALLED = False


"""
Implementation note regarding using tensorstore with num_workers > 0
https://github.com/google/tensorstore/issues/61
TL;DR TensorStore is not Send.
"""


class ZarrTracks(Reader):
    chunked = True

    def __init__(self, name: str, path: Union[str, Path]):
        if not ZARR_TENSORSTORE_INSTALLED:
            raise ImportError(
                "ZarrTracks support requires zarr and tensorstore to be installed."
            )

        self.name = name
        self.path = Path(path)
        z = zarr.open_group(self.path, mode="r")  # pyright: ignore
        self.contigs = cast(Dict[str, int], z.attrs["contigs"])
        self.sizes = cast(Dict[str, int], z.attrs["sizes"])
        self.coords = {d: np.asarray(z.attrs[d]) for d in self.sizes}
        self.samples = cast(Optional[List[str]], z.attrs.get("sample", None))
        self.ploidy = cast(Optional[int], z.attrs.get("ploid", None))
        self.dtype = z[next(iter(self.contigs))].dtype
        # each is (s? p? l)
        self.tstores: Optional[Dict[str, Any]] = None

    def rev_strand_fn(self, x):
        return x[..., ::-1]

    @classmethod
    def from_reader(
        cls,
        reader: Reader,
        path: Union[str, Path],
        chunk_shape: Optional[Tuple[int, ...]] = None,
    ):
        if isinstance(path, str):
            path = Path(path)
        z = zarr.open_group(path)  # pyright: ignore[reportPossiblyUnboundVariable]
        z.attrs["contigs"] = reader.contigs
        z.attrs["sizes"] = reader.sizes
        n_chunks = len(reader.contigs)

        if "sample" in reader.sizes:
            samples = reader.coords["sample"].tolist()
            n_chunks *= len(samples) // 10 + 1
            z.attrs["sample"] = samples
        if "ploid" in reader.sizes:
            z.attrs["ploid"] = reader.sizes["ploidy"]

        if chunk_shape is None:
            if "sample" in reader.sizes and "ploid" in reader.sizes:
                chunk_shape = (10, reader.sizes["ploid"], int(5e5))
            elif "sample" in reader.sizes:
                chunk_shape = (20, int(5e5))
            elif "ploidy" in reader.sizes:
                chunk_shape = (2, int(5e5))

        chunk_layout = ts.ChunkLayout(  # pyright: ignore
            chunk_shape=chunk_shape
        )

        with tqdm(total=n_chunks) as pbar:
            for contig, e in reader.contigs.items():
                ts_open = partial(
                    ts.open,
                    {
                        "driver": "zarr",
                        "kvstore": {"driver": "file", "path": str(path / contig)},
                        "metadata": {
                            "compressor": {
                                "id": "blosc",
                                "shuffle": -1,
                                "clevel": 5,
                                "cname": "lz4",
                            }
                        },
                    },
                    read=True,
                    write=True,
                    create=True,
                    delete_existing=True,
                    chunk_layout=chunk_layout,
                )
                tstore = None
                if "sample" in reader.sizes:
                    for b_idx, samples in enumerate(
                        batched(reader.coords["sample"], 10)
                    ):
                        pbar.set_description(f"Reading {contig}")
                        data = reader.read(contig, 0, e, sample=samples)

                        if tstore is None:
                            tstore = ts_open(
                                dtype=data.dtype,
                                shape=(reader.sizes["sample"], *data.shape[1:]),
                            ).result()

                        pbar.set_description(f"Writing {contig}")
                        start = b_idx * 10
                        stop = b_idx * 10 + len(samples)
                        tstore[start:stop].write(data).result()

                        pbar.update()
                else:
                    pbar.set_description(f"Reading {contig}")
                    data = reader.read(contig, 0, e)

                    if tstore is None:
                        tstore = ts_open(dtype=data.dtype, shape=data.shape).result()

                    pbar.set_description(f"Writing {contig}")
                    tstore.write(data).result()

                    pbar.update()
        zarr.consolidate_metadata(str(path))  # pyright: ignore

        return cls(reader.name, path)

    def _tstore(self, contig: str):
        tstore = ts.open(  # pyright: ignore
            {
                "driver": "zarr",
                "kvstore": {"driver": "file", "path": str(self.path / contig)},
            },
            read=True,
        ).result()
        return tstore

    def read(
        self, contig: str, starts: ArrayLike, ends: ArrayLike, **kwargs
    ) -> NDArray:
        if self.tstores is None:
            self.tstores = {contig: self._tstore(contig) for contig in self.contigs}

        contig = _normalize_contig_name(contig, self.contigs)  # pyright: ignore[reportAssignmentType]
        if contig is None:
            raise ValueError(f"Contig {contig} not found")

        tstore = self.tstores[contig]
        samples = cast(Optional[List[str]], kwargs.get("sample", self.samples))
        if samples is not None:
            if self.samples is None:
                samples = None
            elif missing := set(samples).difference(self.samples):
                raise ValueError(f"Samples {missing} were not found")
            else:
                key_idx, query_idx = np.intersect1d(
                    self.samples, samples, return_indices=True
                )[1:]
                sample_idx = key_idx[query_idx]
                # (s p? l)
                tstore = tstore[sample_idx]

        ploidy = cast(Optional[ArrayLike], kwargs.get("ploid", None))
        if ploidy is not None:
            if self.ploidy is None:
                ploidy = None
            else:
                haplotype_idx = np.asarray(ploidy, dtype=int)
                if (haplotype_idx >= self.ploidy).any():
                    raise ValueError("Ploidies requested exceed maximum ploidy")
                # (s? p l)
                if self.samples is None:
                    # (p l)
                    tstore = tstore[haplotype_idx]
                else:
                    # (s p l)
                    tstore = tstore[:, haplotype_idx]

        starts = np.atleast_1d(np.asarray(starts, dtype=int))
        ends = np.atleast_1d(np.asarray(ends, dtype=int))

        sub_values = [None] * len(starts)
        for i, (s, e) in enumerate(zip(starts, ends)):
            # no variants in query regions
            if s == e:
                continue
            # (s? p? l)
            sub_values[i] = tstore[..., s:e]

        values = ts.concat(sub_values, axis=-1)[  # pyright: ignore  # pyright: ignore
            ts.d[0].translate_to[0]  # pyright: ignore
        ]

        values = cast(NDArray, values.read().result())

        return values

    def vidx(
        self,
        contigs: ArrayLike,
        starts: ArrayLike,
        length: int,
        samples: Optional[ArrayLike] = None,
        ploidy: Optional[ArrayLike] = None,
        **kwargs,
    ) -> NDArray:
        if self.tstores is None:
            self.tstores = {contig: self._tstore(contig) for contig in self.contigs}

        contigs = np.array(
            [
                _normalize_contig_name(c, self.contigs)
                for c in np.atleast_1d(np.asarray(contigs))
            ]
        )
        if (contigs == None).any():  # noqa: E711
            raise ValueError("Some contigs not found")

        if samples is not None:
            if self.samples is None:
                raise ValueError("No sample information available")

            unique_samples, inverse = np.unique(samples, return_inverse=True)
            if missing := set(unique_samples).difference(self.samples):
                raise ValueError(f"Samples {missing} were not found")

            key_idx, query_idx = np.intersect1d(
                self.samples, unique_samples, return_indices=True, assume_unique=True
            )[1:]
            sample_idx = key_idx[query_idx[inverse]]
        else:
            if self.samples is None:
                sample_idx = [None]
            else:
                sample_idx = [slice(None)]

        if ploidy is not None:
            if self.ploidy is None:
                raise ValueError("No ploidy information available")

            haplotype_idx = np.asarray(ploidy, dtype=int)
            if (haplotype_idx >= self.ploidy).any():
                raise ValueError("Ploidies requested exceed maximum ploidy")
        else:
            if self.ploidy is None:
                haplotype_idx = [None]
            else:
                haplotype_idx = [slice(None)]

        starts = np.atleast_1d(np.asarray(starts, dtype=int))
        ends = starts + length

        sub_values = [None] * len(starts)
        for i, (c, s, e, sp, h) in enumerate(
            zip_longest(contigs, starts, ends, sample_idx, haplotype_idx)
        ):
            # no variants in query regions
            if s == e:
                continue
            tstore = self.tstores[c]
            # (s? p? l)
            if sp is None and h is None:
                # (l)
                sub_values[i] = tstore[s:e]
            elif sp is None:
                # (p l)
                sub_values[i] = tstore[h, s:e]
            elif h is None:
                # (s p l)
                sub_values[i] = tstore[sp, s:e]
            else:
                sub_values[i] = tstore[sp, h, s:e]

        values = ts.concat(sub_values, axis=-1)[  # pyright: ignore
            ts.d[0].translate_to[0]  # pyright: ignore
        ]

        values = cast(NDArray, values.read().result())

        return values



from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Union,
    cast,
)

import numpy as np
from attrs import define
from loguru import logger
from numpy.typing import NDArray

from .._fasta import Fasta
from .._utils import _normalize_contig_name

__all__ = []


@define
class Reference:
    reference: NDArray[np.uint8]
    contigs: List[str]
    offsets: NDArray[np.uint64]
    pad_char: int

    @classmethod
    def from_path_and_contigs(cls, fasta: Union[str, Path], contigs: List[str]):
        _fasta = Fasta("ref", fasta, "N")

        if not _fasta.cache_path.exists():
            logger.info("Memory-mapping FASTA file for faster access.")
            _fasta._write_to_cache()

        contigs = cast(
            List[str],
            [_normalize_contig_name(c, _fasta.contigs) for c in contigs],
        )
        _fasta.sequences = _fasta._get_sequences(contigs)
        if TYPE_CHECKING:
            assert _fasta.sequences is not None
            assert _fasta.pad is not None
        refs: List[NDArray[np.bytes_]] = []
        next_offset = 0
        _ref_offsets: Dict[str, int] = {}
        for contig in contigs:
            arr = _fasta.sequences[contig]
            refs.append(arr)
            _ref_offsets[contig] = next_offset
            next_offset += len(arr)
        reference = np.concatenate(refs).view(np.uint8)
        pad_char = ord(_fasta.pad)
        if any(c is None for c in contigs):
            raise ValueError("Contig names in metadata do not match reference.")
        ref_offsets = np.empty(len(contigs) + 1, np.uint64)
        ref_offsets[:-1] = np.array([_ref_offsets[c] for c in contigs], dtype=np.uint64)
        ref_offsets[-1] = len(reference)
        return cls(reference, contigs, ref_offsets, pad_char)



import numpy as np
from genvarloader._dataset._utils import splits_sum_le_value


def test_splits_sum_le_value():
    max_size = 10
    sizes = np.array([3, 5, 2, 4, 7, 5, 2], np.int32)
    splits = splits_sum_le_value(sizes, max_size)
    np.testing.assert_equal(splits, np.array([0, 3, 4, 5, 7], np.intp))
    np.testing.assert_array_less(np.add.reduceat(sizes, splits[:-1]), max_size + 1)



from __future__ import annotations

from typing import Any, Generic, Optional, Tuple, TypeGuard, TypeVar, Union, cast

import awkward as ak
import numba as nb
import numpy as np
from attrs import define
from awkward.contents import ListOffsetArray, NumpyArray, RegularArray
from awkward.index import Index64
from einops import repeat
from numpy.typing import NDArray

from ._types import DTYPE, AnnotatedHaps, Idx
from ._utils import _lengths_to_offsets, idx_like_to_array

__all__ = ["Ragged", "RaggedIntervals", "INTERVAL_DTYPE", "pad_ragged"]

RDTYPE = TypeVar("RDTYPE", bound=np.generic)


@define
class RaggedAnnotatedHaps:
    haps: Ragged[np.bytes_]
    var_idxs: Ragged[np.int32]
    ref_coords: Ragged[np.int32]

    @property
    def shape(self):
        return self.haps.shape

    def to_padded(self) -> AnnotatedHaps:
        haps = self.haps.to_padded(b"N")
        var_idxs = self.var_idxs.to_padded(-1)
        ref_coords = self.ref_coords.to_padded(-1)
        return AnnotatedHaps(haps, var_idxs, ref_coords)

    def reshape(self, shape: tuple[int, ...]) -> RaggedAnnotatedHaps:
        return RaggedAnnotatedHaps(
            self.haps.reshape(shape),
            self.var_idxs.reshape(shape),
            self.ref_coords.reshape(shape),
        )

    def squeeze(self, axis: int | tuple[int, ...] | None = None) -> RaggedAnnotatedHaps:
        return RaggedAnnotatedHaps(
            self.haps.squeeze(axis),
            self.var_idxs.squeeze(axis),
            self.ref_coords.squeeze(axis),
        )

    def to_fixed_shape(self, shape: tuple[int, ...]) -> AnnotatedHaps:
        haps = self.haps.data.reshape(shape)
        var_idxs = self.var_idxs.data.reshape(shape)
        ref_coords = self.ref_coords.data.reshape(shape)
        return AnnotatedHaps(haps, var_idxs, ref_coords)


def is_rag_dtype(rag: Ragged, dtype: type[DTYPE]) -> TypeGuard[Ragged[DTYPE]]:
    return np.issubdtype(rag.data.dtype, dtype)


@define
class Ragged(Generic[RDTYPE]):
    """Ragged array i.e. a rectilinear array where the final axis is ragged. Should not be initialized
    directly, use :meth:`from_offsets()` or :meth:`from_lengths()` instead.
    Does not have a :code:`__getitem__` method currently.

    Examples
    --------

    .. code-block:: python

        r = Ragged.from_lengths(np.arange(10), np.array([3, 2, 5]))
        assert r.offsets ==  np.array([0, 3, 5, 10])
        assert r.data[r.offsets[0]:r.offsets[1]] == np.array([0, 1, 2])
        assert r.data[r.offsets[1]:r.offsets[2]] == np.array([3, 4])
        assert r.data[r.offsets[2]:r.offsets[3]] == np.array([5, 6, 7, 8, 9])

    """

    data: NDArray[RDTYPE]
    """A 1D array of the data."""
    shape: Tuple[int, ...]
    """Shape of the ragged array, excluding the length dimension. For example, if
        the shape is (2, 3), then the j, k-th element can be mapped to an index for
        offsets with :code:`i = np.ravel_multi_index((j, k), shape)`. The number of ragged
        elements corresponds to the product of the shape."""
    maybe_offsets: Optional[NDArray[np.int64]] = None
    maybe_lengths: Optional[NDArray[np.int32]] = None

    def __attrs_post_init__(self):
        if self.maybe_offsets is None and self.maybe_lengths is None:
            raise ValueError("Either offsets or lengths must be provided.")

    def __len__(self):
        return self.shape[0]

    def item(self):
        a = self.squeeze()
        if a.shape != ():
            raise ValueError("Array has more than 1 ragged element.")
        return a.data

    @property
    def ndim(self) -> int:
        """Number of dimensions of the ragged array."""
        return len(self.shape)

    @property
    def offsets(self) -> NDArray[np.int64]:
        """Offsets into the data array to get corresponding elements. The i-th element
        is accessible as :code:`data[offsets[i]:offsets[i+1]]`."""
        if self.maybe_offsets is None:
            self.maybe_offsets = _lengths_to_offsets(self.lengths)
        return self.maybe_offsets

    @property
    def lengths(self) -> NDArray[np.int32]:
        """Array with appropriate shape containing lengths of each element in the ragged array."""
        if self.maybe_lengths is None:
            self.maybe_lengths = np.diff(self.offsets).reshape(self.shape)
        return self.maybe_lengths

    @staticmethod
    def from_offsets(
        data: NDArray[DTYPE],
        shape: Union[int, Tuple[int, ...]],
        offsets: NDArray[np.int64],
    ) -> "Ragged[DTYPE]":
        """Create a Ragged array from data and offsets.

        Parameters
        ----------
        data
            1D data array.
        shape
            Shape of the ragged array, excluding the length dimension.
        offsets
            Offsets into the data array to get corresponding elements.
        """
        if isinstance(shape, int):
            shape = (shape,)
        return Ragged(data, shape, maybe_offsets=offsets)

    @staticmethod
    def from_lengths(
        data: NDArray[DTYPE], lengths: NDArray[np.int32]
    ) -> "Ragged[DTYPE]":
        """Create a Ragged array from data and lengths. The lengths array should have
        the intended shape of the Ragged array.

        Parameters
        ----------
        data
            1D data array.
        lengths
            Lengths of each element in the ragged array.
        """
        return Ragged(data, lengths.shape, maybe_lengths=lengths)

    @staticmethod
    def concat(*arrays: "Ragged[DTYPE]", axis: int) -> "Ragged[DTYPE]":
        # """Concatenate multiple Ragged arrays along a given axis."""
        # need to check whether this would lead to incorrect indexing
        raise NotImplementedError
        if len(set((*a.shape[:axis], *a.shape[axis + 1 :]) for a in arrays)) != 1:
            raise ValueError(
                f"All arrays must have the same shape except along axis {axis}."
            )

        if len(set(a.data.dtype for a in arrays)) != 1:
            raise ValueError("All arrays must have the same dtype.")

        data = np.concatenate([a.data for a in arrays])
        lengths = np.concatenate([a.lengths for a in arrays], axis=axis)
        return Ragged.from_lengths(data, lengths)

    @staticmethod
    def stack(*arrays: "Ragged[DTYPE]") -> "Ragged[DTYPE]":
        """Stack multiple ragged arrays along a new first axis."""
        if len(set(a.shape for a in arrays)) != 1:
            raise ValueError("All arrays must have the same shape.")

        if len(set(a.data.dtype for a in arrays)) != 1:
            raise ValueError("All arrays must have the same dtype.")

        data = np.concatenate([a.data for a in arrays])
        lengths = np.stack([a.lengths for a in arrays], axis=0)
        return Ragged.from_lengths(data, lengths)

    def to_padded(self, pad_value: Any) -> NDArray[RDTYPE]:
        """Convert this Ragged array to a rectilinear array by right-padding each entry with a value.
        The final axis will have the maximum length across all entries.

        Parameters
        ----------
        pad_value
            Value to pad the entries with.

        Returns
        -------
            Padded array with shape :code:`(*self.shape, self.lengths.max())`.
        """
        length = self.lengths.max()
        shape = (*self.shape, length)
        if self.data.dtype.str == "|S1":
            if isinstance(pad_value, (str, bytes)):
                if len(pad_value) != 1:
                    raise ValueError(
                        "Tried padding an S1 array with a `pad_value` that is multiple characters."
                    )
                pad_value = np.uint8(ord(pad_value))
            elif isinstance(pad_value, int):
                if pad_value < 0 or pad_value > 255:
                    raise ValueError(
                        "Tried padding an S1 array with an integer `pad_value` outside the ASCII range."
                    )
                pad_value = np.uint8(pad_value)
            else:
                raise ValueError(
                    "Tried padding an S1 array with a `pad_value` that isn't a string, byte, or integer."
                )
            padded = np.empty((np.prod(shape[:-1]), shape[-1]), dtype=np.uint8)
            pad_ragged(self.data.view(np.uint8), self.offsets, pad_value, padded)
            padded = padded.view(self.data.dtype).reshape(shape)
        else:
            padded = np.empty((np.prod(shape[:-1]), shape[-1]), dtype=self.data.dtype)
            pad_ragged(self.data, self.offsets, pad_value, padded)
            padded = padded.reshape(shape)

        return padded

    def squeeze(
        self, axis: Optional[Union[int, Tuple[int, ...]]] = None
    ) -> Ragged[RDTYPE]:
        """Squeeze the ragged array along the given non-ragged axis."""
        return Ragged.from_lengths(self.data, self.lengths.squeeze(axis))

    def reshape(self, shape: Tuple[int, ...]) -> Ragged[RDTYPE]:
        """Reshape non-ragged axes."""
        # this is correct because all reshaping operations preserve the layout i.e. raveled ordered
        return Ragged.from_lengths(self.data, self.lengths.reshape(shape))

    def __str__(self):
        return (
            f"Ragged<shape={self.shape} dtype={self.data.dtype} size={self.data.size}>"
        )

    def __getitem__(self, idx: Union[Idx, Tuple[Idx, ...]]):
        if not isinstance(idx, tuple):
            idx = (idx,)

        if len(idx) < len(self.shape):
            raise IndexError(
                f"Too few indices for array: expected {len(self.shape)}, got {len(idx)}."
            )

        idx = tuple(
            idx_like_to_array(_idx, self.shape[i]) for i, _idx in enumerate(idx)
        )
        idx = np.ravel_multi_index(idx, self.shape).squeeze()

        return self.data[self.offsets[idx] : self.offsets[idx + 1]]

    def to_awkward(self) -> ak.Array:
        """Convert to an `Awkward <https://awkward-array.org/doc/main/>`_ array without copying. Note that this effectively
        returns a view of the data, so modifying the data will modify the original array."""
        layout = ListOffsetArray(
            Index64(self.offsets),
            NumpyArray(self.data),  # type: ignore | NDArray[RDTYPE] is ArrayLike
        )

        for size in reversed(self.shape[1:]):
            layout = RegularArray(layout, size)

        return ak.Array(layout)

    @classmethod
    def from_awkward(cls, awk: "ak.Array") -> "Ragged":
        """Convert from an `Awkward <https://awkward-array.org/doc/main/>`_ array without copying. Note that this effectively
        returns a view of the data, so modifying the data will modify the original array."""
        # parse shape
        shape_str = awk.typestr.split(" * ")
        try:
            shape = tuple(map(int, shape_str[:-2]))
        except ValueError as err:
            raise ValueError(
                f"Only the final axis of an awkward array may be variable to convert to ragged, but got {awk.type}."
            ) from err

        # extract data and offsets
        data = ak.flatten(awk, axis=None).to_numpy()
        layout = awk.layout
        while hasattr(layout, "content"):
            if isinstance(layout, ListOffsetArray):
                offsets = layout.offsets.data
                offsets = cast(NDArray[np.int64], offsets)
                rag = cls.from_offsets(data, shape, offsets)
                break
            else:
                layout = layout.content
        else:
            lengths = ak.count(awk, axis=-1).to_numpy()
            rag = cls.from_lengths(data, lengths)

        return rag


INTERVAL_DTYPE = np.dtype(
    [("start", np.int32), ("end", np.int32), ("value", np.float32)], align=True
)
RaggedIntervals = Ragged[np.void]


@nb.njit(parallel=True, nogil=True, cache=True)
def pad_ragged(
    data: NDArray[DTYPE],
    offsets: NDArray[np.integer],
    pad_value: DTYPE,
    out: NDArray[DTYPE],
):
    for i in nb.prange(len(offsets) - 1):
        start, end = offsets[i], offsets[i + 1]
        entry_len = end - start
        out[i, :entry_len] = data[start:end]
        out[i, entry_len:] = pad_value
    return out


NUCLEOTIDES = b"ACGT"
COMPLEMENTS = b"TGCA"


@nb.njit(parallel=True, nogil=True, cache=True)
def _rc_helper(
    data: NDArray[np.uint8], offsets: NDArray[np.int64], mask: NDArray[np.bool_]
) -> NDArray[np.uint8]:
    out = data.copy()
    for i in nb.prange(len(offsets) - 1):
        start, end = offsets[i], offsets[i + 1]
        _data = data[start:end]
        _out = out[start:end]
        if mask[i]:
            for nuc, comp in zip(NUCLEOTIDES, COMPLEMENTS):
                _out[_data == nuc] = comp
            _out[:] = _out[::-1]
        else:
            _out[:] = _data
    return out


def _reverse_complement(
    seqs: Ragged[np.bytes_], mask: NDArray[np.bool_]
) -> Ragged[np.bytes_]:
    # (b [p] ~l), (b)
    if seqs.ndim == 2:
        ploidy = seqs.shape[1]
        mask = repeat(mask, "b -> (b p)", p=ploidy)
    rc_seqs = _rc_helper(seqs.data.view(np.uint8), seqs.offsets, mask)
    return Ragged.from_offsets(rc_seqs.view("S1"), seqs.shape, seqs.offsets)


@nb.njit(parallel=True, nogil=True, cache=True)
def _reverse_helper(data: NDArray, offsets: NDArray[np.int64], mask: NDArray[np.bool_]):
    for i in nb.prange(len(offsets) - 1):
        if mask[i]:
            start, end = offsets[i], offsets[i + 1]
            if start > 0:
                data[start:end] = data[end - 1 : start - 1 : -1]
            else:
                data[start:end] = data[end - 1 :: -1]


def _reverse(tracks: Ragged, mask: NDArray[np.bool_]):
    """Reverses data along the ragged axis in-place."""
    # (b t [p] ~l), (b)
    if tracks.ndim == 2:
        n_tracks = tracks.shape[1]
        mask = repeat(mask, "b -> (b t)", t=n_tracks)
    elif tracks.ndim == 3:
        n_tracks, ploidy = tracks.shape[1:]
        mask = repeat(mask, "b -> (b t p)", t=n_tracks, p=ploidy)
    _reverse_helper(tracks.data, tracks.offsets, mask)


def _jitter(
    *arrays: Ragged,
    max_jitter: int,
    seed: Optional[Union[int, np.random.Generator]] = None,
    starts: NDArray[np.int64] | None = None,
):
    """Jitter the data along the ragged axis by up to `max_jitter`.

    Assumes only the first axis is to be jittered. In other words, assumes that no user would want to
    jitter independently across ploidy or tracks.
    """
    batch_size = arrays[0].shape[0]

    if starts is None:
        if not isinstance(seed, np.random.Generator):
            seed = np.random.default_rng(seed)
        starts = seed.integers(0, 2 * max_jitter + 1, size=batch_size)
    out_offsets = tuple(_lengths_to_offsets(a.lengths - 2 * max_jitter) for a in arrays)

    jittered_data = _jitter_helper(
        tuple(a.data for a in arrays),
        tuple(a.offsets for a in arrays),
        tuple(a.shape for a in arrays),
        tuple(out_offsets),
        starts,
    )

    out = tuple(
        Ragged.from_offsets(jit_data, rag_arr.shape, out_offsets[i])
        for i, (rag_arr, jit_data) in enumerate(zip(arrays, jittered_data))
    )

    return out


@nb.njit(parallel=True, nogil=True, cache=True)
def _jitter_helper(
    data: Tuple[NDArray, ...],
    offsets: Tuple[NDArray[np.int64], ...],
    shapes: Tuple[Tuple[int, ...], ...],
    out_offsets: Tuple[NDArray[np.int64], ...],
    starts: NDArray[np.int64],
) -> Tuple[NDArray, ...]:
    """Helper to jitter ragged data. Ragged arrays should have shape (batch, ...).

    Parameters
    ----------
    data
        Tuple of ragged data arrays.
    offsets
        Tuple of offsets arrays.
    out_offsets
        Tuple of output offsets arrays.
    starts
        Shape: (batch). Array of starting points for jittering.
    """
    n_arrays = len(data)
    batch_size = len(starts)
    out_data = tuple(
        np.empty(out_offsets[i][-1], dtype=d.dtype) for i, d in enumerate(data)
    )
    for arr in nb.prange(n_arrays):
        arr_data = data[arr]
        arr_offsets = offsets[arr]

        out_arr_data = out_data[arr]
        out_arr_offsets = out_offsets[arr]

        arr_shape = shapes[arr]
        n_per_batch = np.prod(arr_shape[1:])

        for jit in nb.prange(batch_size):
            idx_s = jit * n_per_batch
            idx_e = (jit + 1) * n_per_batch

            jit_s = starts[jit]

            for row in nb.prange(idx_s, idx_e):
                row_s = arr_offsets[row]
                out_row_s, out_row_e = out_arr_offsets[row], out_arr_offsets[row + 1]
                out_len = out_row_e - out_row_s
                out_arr_data[out_row_s:out_row_e] = arr_data[
                    row_s + jit_s : row_s + jit_s + out_len
                ]

    return out_data



from pathlib import Path

import pyBigWig
import typer


def main(n_samples: int = 2):
    data_dir = Path(__file__).resolve().parent
    for i in range(n_samples):
        path = data_dir / f"sample_{i}.bw"
        with pyBigWig.open(str(path), "w") as bw:
            bw.addHeader([("chr1", 2000), ("chr2", 1000)], maxZooms=0)
            bw.addEntries(["chr1", "chr1"], [1, 100], [5, 105], [1.0, 2.0])
            bw.addEntries(["chr2", "chr2"], [1, 100], [5, 105], [1.0, 2.0])


if __name__ == "__main__":
    typer.run(main)



# %%
import numpy as np
from genvarloader._dataset._genotypes import (
    SparseGenotypes,
    reconstruct_haplotype_from_sparse,
)
from pytest_cases import parametrize_with_cases


# %%
def case_snps():
    positions = np.array([1, 3], np.int32)
    sizes = np.zeros(2, dtype=np.int32)

    # (s p v)
    genos = np.array([[[0, 1]]], dtype=np.int8)
    first_v_idxs = np.array([0], dtype=np.int32)
    offsets = np.array([0, 2], np.int64)

    shift = 0
    alt_alleles = np.frombuffer(b"A" + b"T", dtype=np.uint8)
    alt_offsets = np.array([0, 1, 2], dtype=np.uintp)
    ref = np.frombuffer(b"ACGG", dtype=np.uint8)
    ref_start = 1

    desired = np.frombuffer(b"CGT", dtype="S1")
    annot_v_idxs = np.array([-1, -1, 1], dtype=np.int32)
    annot_pos = np.array([1, 2, 3], dtype=np.int32)

    sparse_genos = SparseGenotypes.from_dense(
        genos=genos, first_v_idxs=first_v_idxs, offsets=offsets
    )

    return (
        positions,
        sizes,
        shift,
        alt_alleles,
        alt_offsets,
        ref,
        sparse_genos,
        ref_start,
        desired,
        annot_v_idxs,
        annot_pos,
    )


def case_indels():
    positions = np.array([1, 3], np.int32)
    sizes = np.array([-1, 1], dtype=np.int32)

    # (s p v) : (1 1 2)
    genos = np.array([[[1, 1]]], dtype=np.int8)
    first_v_idxs = np.array([0], dtype=np.int32)
    offsets = np.array([0, 2], np.int64)

    shift = 0
    alt_alleles = np.frombuffer(b"G" + b"AT", dtype=np.uint8)
    alt_offsets = np.array([0, 1, 3], np.uintp)
    ref = np.frombuffer(b"ACGG", dtype=np.uint8)
    ref_start = 0

    desired = np.frombuffer(b"AGAT", dtype="S1")
    annot_v_idxs = np.array([-1, 0, 1, 1], dtype=np.int32)
    annot_pos = np.array([0, 1, 3, 3], dtype=np.int32)

    sparse_genos = SparseGenotypes.from_dense(
        genos=genos, first_v_idxs=first_v_idxs, offsets=offsets
    )

    return (
        positions,
        sizes,
        shift,
        alt_alleles,
        alt_offsets,
        ref,
        sparse_genos,
        ref_start,
        desired,
        annot_v_idxs,
        annot_pos,
    )


def case_spanning_del_pad():
    positions = np.array([0], np.int32)
    sizes = np.array([-1], dtype=np.int32)

    # (s p v) : (1 1 2)
    genos = np.array([[[1]]], dtype=np.int8)
    first_v_idxs = np.array([0], dtype=np.int32)
    offsets = np.array([0, 1], np.int64)

    shift = 0
    alt_alleles = np.frombuffer(b"G", dtype=np.uint8)
    alt_offsets = np.array([0, 1], np.uintp)
    ref = np.frombuffer(b"ACGG", dtype=np.uint8)
    ref_start = 1

    desired = np.frombuffer(b"GGN", dtype="S1")
    annot_v_idxs = np.array([-1, -1, -1], dtype=np.int32)
    annot_pos = np.array([2, 3, -1], dtype=np.int32)

    sparse_genos = SparseGenotypes.from_dense(
        genos=genos, first_v_idxs=first_v_idxs, offsets=offsets
    )

    return (
        positions,
        sizes,
        shift,
        alt_alleles,
        alt_offsets,
        ref,
        sparse_genos,
        ref_start,
        desired,
        annot_v_idxs,
        annot_pos,
    )


def case_shift_ins():
    positions = np.array([1, 3], np.int32)
    sizes = np.array([1, 1], dtype=np.int32)

    # (s p v) : (1 1 2)
    genos = np.array([[[1, 1]]], dtype=np.int8)
    first_v_idxs = np.array([0], dtype=np.int32)
    offsets = np.array([0, 2], np.int64)

    shift = 1
    alt_alleles = np.frombuffer(b"TC" + b"GA", dtype=np.uint8)
    alt_offsets = np.array([0, 2, 4], np.uintp)
    ref = np.frombuffer(b"ACGG", dtype=np.uint8)
    ref_start = 0

    desired = np.frombuffer(b"TCGG", dtype="S1")
    annot_v_idxs = np.array([0, 0, -1, 1], dtype=np.int32)
    annot_pos = np.array([1, 1, 2, 3], dtype=np.int32)

    sparse_genos = SparseGenotypes.from_dense(
        genos=genos, first_v_idxs=first_v_idxs, offsets=offsets
    )

    return (
        positions,
        sizes,
        shift,
        alt_alleles,
        alt_offsets,
        ref,
        sparse_genos,
        ref_start,
        desired,
        annot_v_idxs,
        annot_pos,
    )


@parametrize_with_cases(
    "positions, sizes, shift, alt_alleles, alt_offsets, ref, sparse_genos, ref_start, desired, annot_v_idxs, annot_pos",
    cases=".",
)
def test_sparse(
    positions,
    sizes,
    shift,
    alt_alleles,
    alt_offsets,
    ref,
    sparse_genos: SparseGenotypes,
    ref_start,
    desired,
    annot_v_idxs,
    annot_pos,
):
    offset_idx = 0
    actual = np.empty(len(ref) - ref_start, np.uint8)
    actual_annot_v_idxs = np.empty(len(ref) - ref_start, np.int32)
    actual_annot_pos = np.empty(len(ref) - ref_start, np.int32)
    reconstruct_haplotype_from_sparse(
        offset_idx=offset_idx,
        geno_v_idxs=sparse_genos.variant_idxs,
        geno_offsets=sparse_genos.offsets,
        positions=positions,
        sizes=sizes,
        shift=shift,
        alt_alleles=alt_alleles,
        alt_offsets=alt_offsets,
        ref=ref,
        ref_start=ref_start,
        out=actual,
        pad_char=ord(b"N"),
        annot_v_idxs=actual_annot_v_idxs,
        annot_ref_pos=actual_annot_pos,
    )

    np.testing.assert_equal(actual.view("S1"), desired)
    np.testing.assert_equal(actual_annot_v_idxs, annot_v_idxs)
    np.testing.assert_equal(actual_annot_pos, annot_pos)


# %%
# test_sparse(*case_snps())
# test_sparse(*case_indels())
# test_sparse(*case_spanning_del_pad())
# test_sparse(*case_shift_ins())



from __future__ import annotations

from pathlib import Path
from typing import Dict, List, Optional, Union

import numpy as np
import polars as pl
import pyBigWig
from numpy.typing import ArrayLike, NDArray

from ._ragged import INTERVAL_DTYPE, RaggedIntervals
from ._types import Reader
from ._utils import _get_rel_starts, _lengths_to_offsets, _normalize_contig_name
from .genvarloader import count_intervals as bw_count_intervals
from .genvarloader import intervals as bw_intervals

__all__ = []


class BigWigs(Reader):
    dtype = np.float32  # pyBigWig always returns f32
    chunked = False

    def __init__(self, name: str, paths: Dict[str, str]) -> None:
        """Read data from bigWig files.

        Parameters
        ----------
        name
            Name of the reader, for example `'signal'`.
        paths
            Dictionary of sample names and paths to bigWig files for those samples.
        """
        self.name = name
        self.paths = paths
        self.readers = None
        self.samples = list(self.paths)
        self.coords = {"sample": np.asarray(self.samples)}
        self.sizes = {"sample": len(self.samples)}
        contigs: Optional[Dict[str, int]] = None
        for path in self.paths.values():
            with pyBigWig.open(path, "r") as f:
                if contigs is None:
                    contigs = {
                        contig: int(length) for contig, length in f.chroms().items()
                    }
                else:
                    common_contigs = contigs.keys() & f.chroms()
                    contigs = {k: v for k, v in contigs.items() if k in common_contigs}
        if contigs is None:
            raise ValueError("No bigWig files provided.")
        self.contigs: Dict[str, int] = contigs

    @classmethod
    def from_table(cls, name: str, table: Union[str, Path, pl.DataFrame]):
        """Read data from bigWig files.

        Parameters
        ----------
        name
            Name of the reader, for example `'signal'`.
        table
            Path to a table or a DataFrame containing sample names and paths to bigWig files for those samples.
            It must have columns "sample" and "path".
        """
        if isinstance(table, (str, Path)):
            table = Path(table)
            if table.suffix == ".csv":
                table = pl.read_csv(table)
            elif table.suffix == ".tsv" or table.suffix == ".txt":
                table = pl.read_csv(table, separator="\t")
            else:
                raise ValueError("Table should be a csv or tsv file.")
        paths = dict(zip(table["sample"], table["path"]))
        return cls(name, paths)

    def rev_strand_fn(self, x):
        return x[..., ::-1]

    def close(self):
        if self.readers is not None:
            for reader in self.readers.values():
                reader.close()
            self.readers = None

    def read(
        self,
        contig: str,
        starts: ArrayLike,
        ends: ArrayLike,
        **kwargs,
    ) -> NDArray[np.float32]:
        """Read data corresponding to given genomic coordinates. The output shape will
        have length as the final dimension/axis i.e. (..., length).

        Parameters
        ----------
        contig
            Name of the contig/chromosome.
        starts : ArrayLike
            Start coordinates, 0-based.
        ends : ArrayLike
            End coordinates, 0-based, exclusive.
        sample
            Name of the samples to read data from.

        Returns
        -------
            Shape: (samples length). Data corresponding to the given genomic coordinates and samples.
        """
        _contig = _normalize_contig_name(contig, self.contigs)
        if _contig is None:
            raise ValueError(f"Contig {contig} not found.")
        else:
            contig = _contig

        if self.readers is None:
            self.readers = {s: pyBigWig.open(p, "r") for s, p in self.paths.items()}

        samples = kwargs.get("sample", self.samples)
        if isinstance(samples, str):
            samples = [samples]
        if not set(samples).issubset(self.samples):
            raise ValueError(f"Sample {samples} not found in bigWig paths.")

        starts = np.atleast_1d(np.asarray(starts, dtype=int))
        ends = np.atleast_1d(np.asarray(ends, dtype=int))
        rel_starts = _get_rel_starts(starts, ends)
        rel_ends = rel_starts + (ends - starts)

        out = np.empty((len(samples), (ends - starts).sum()), dtype=np.float32)
        for s, e, r_s, r_e in zip(starts, ends, rel_starts, rel_ends):
            for i, sample in enumerate(samples):
                out[i, r_s:r_e] = self.readers[sample].values(contig, s, e, numpy=True)

        out = np.nan_to_num(out, copy=False)

        return out

    def count_intervals(
        self,
        contig: str,
        starts: ArrayLike,
        ends: ArrayLike,
        sample: Optional[Union[str, List[str]]] = None,
        **kwargs,
    ) -> NDArray[np.int32]:
        """Count the number of intervals corresponding to given genomic coordinates.

        Parameters
        ----------
        contig
            Name of the contig/chromosome.
        starts : ArrayLike
            Start coordinates, 0-based.
        ends : ArrayLike
            End coordinates, 0-based, exclusive.
        sample
            Name of the samples to read data from.

        Returns
        -------
            Shape: (regions, samples). Number of intervals that overlap with each region and sample.
        """
        _contig = _normalize_contig_name(contig, self.contigs)
        if _contig is None:
            raise ValueError(f"Contig {contig} not found.")
        else:
            contig = _contig

        if sample is None:
            samples = self.samples
        elif isinstance(sample, str):
            samples = [sample]
        else:
            samples = sample

        if not set(samples).issubset(self.samples):
            raise ValueError(f"Sample {samples} not found in bigWig paths.")

        starts = np.atleast_1d(np.asarray(starts, dtype=np.int32))
        ends = np.atleast_1d(np.asarray(ends, dtype=np.int32))
        paths = [self.paths[s] for s in samples]

        # (regions, samples)
        n_per_query = bw_count_intervals(paths, contig, starts, ends)

        return n_per_query

    def intervals(
        self,
        contig: str,
        starts: ArrayLike,
        ends: ArrayLike,
        sample: Optional[Union[str, List[str]]] = None,
        **kwargs,
    ) -> RaggedIntervals:
        """Read intervals corresponding to given genomic coordinates. The output data
        will be a 2D Ragged array of :code:`struct{start, end, value}` with shape (regions, samples).

        Parameters
        ----------
        contig
            Name of the contig/chromosome.
        starts : ArrayLike
            Start coordinates, 0-based.
        ends : ArrayLike
            End coordinates, 0-based, exclusive.
        sample
            Name of the samples to read data from.
        """
        _contig = _normalize_contig_name(contig, self.contigs)
        if _contig is None:
            raise ValueError(f"Contig {contig} not found.")
        else:
            contig = _contig

        if sample is None:
            samples = self.samples
        elif isinstance(sample, str):
            samples = [sample]
        else:
            samples = sample

        if not set(samples).issubset(self.samples):
            raise ValueError(f"Sample {samples} not found in bigWig paths.")

        starts = np.atleast_1d(np.asarray(starts, dtype=np.int32))
        ends = np.atleast_1d(np.asarray(ends, dtype=np.int32))
        paths = [self.paths[s] for s in samples]

        # (regions, samples)
        n_per_query = bw_count_intervals(paths, contig, starts, ends)
        offsets = _lengths_to_offsets(n_per_query.ravel())

        intervals = self._intervals_from_offsets(contig, starts, ends, offsets, sample)

        return intervals

    def _intervals_from_offsets(
        self,
        contig: str,
        starts: ArrayLike,
        ends: ArrayLike,
        offsets: NDArray[np.int64],
        sample: Optional[Union[str, List[str]]] = None,
        **kwargs,
    ) -> RaggedIntervals:
        """This function is unsafe! Reads intervals corresponding to given genomic coordinates
        using provided offsets. If the offsets are incorrect this function is undefined behavior.
        To ensure offsets are correct, use the count_intervals function (see :meth:`intervals()`).
        The output data will be a 2D Ragged array of :code:`struct{start, end, value}` with shape (regions, samples).

        Parameters
        ----------
        contig
            Name of the contig/chromosome.
        starts : ArrayLike
            Start coordinates, 0-based.
        ends : ArrayLike
            End coordinates, 0-based, exclusive.
        offsets
            Offsets corresponding to the returned interval data of shape (regions, samples). Can be
            computed from the number of intervals per query, e.g. with the count_intervals function.
        sample
            Name of the samples to read data from.
        """
        _contig = _normalize_contig_name(contig, self.contigs)
        if _contig is None:
            raise ValueError(f"Contig {contig} not found.")
        else:
            contig = _contig

        if sample is None:
            samples = self.samples
        elif isinstance(sample, str):
            samples = [sample]
        else:
            samples = sample

        if not set(samples).issubset(self.samples):
            raise ValueError(f"Sample {samples} not found in bigWig paths.")

        starts = np.atleast_1d(np.asarray(starts, dtype=np.int32))
        ends = np.atleast_1d(np.asarray(ends, dtype=np.int32))
        paths = [self.paths[s] for s in samples]

        # (intervals, 2) (intervals)
        coordinates, values = bw_intervals(paths, contig, starts, ends, offsets)

        coordinates = coordinates.astype(np.int32)

        intervals = np.empty(len(coordinates), dtype=INTERVAL_DTYPE)
        intervals["start"] = coordinates[:, 0]
        intervals["end"] = coordinates[:, 1]
        intervals["value"] = values

        n_regions = len(starts)
        n_samples = len(samples)
        intervals = RaggedIntervals.from_offsets(
            intervals, (n_regions, n_samples), offsets
        )

        return intervals



from typing import List, Literal, Optional, Sequence, cast

import numpy as np
from attrs import define, evolve
from hirola import HashTable
from more_itertools import collapse
from numpy.typing import NDArray

from genvarloader._dataset._utils import oidx_to_raveled_idx
from genvarloader._types import Idx, StrIdx
from genvarloader._utils import idx_like_to_array, is_dtype


@define
class DatasetIndexer:
    full_region_idxs: NDArray[np.integer]
    """Full map from input region indices to on-disk region indices."""
    full_sample_idxs: NDArray[np.integer]
    """Full map from input sample indices to on-disk sample indices."""
    i2d_map: NDArray[np.integer]
    """Map from input indices to on-disk dataset indices."""
    d2i_map: NDArray[np.integer]
    """Map from on-disk dataset indices to input indices."""
    s2i_map: HashTable
    """Map from input sample names to on-disk sample indices."""
    region_subset_idxs: Optional[NDArray[np.integer]] = None
    """Which input regions are included in the subset."""
    sample_subset_idxs: Optional[NDArray[np.integer]] = None
    """Which input samples are included in the subset."""

    @classmethod
    def from_region_and_sample_idxs(
        cls,
        r_idxs: NDArray[np.integer],
        s_idxs: NDArray[np.integer],
        samples: List[str],
    ):
        shape = len(r_idxs), len(s_idxs)
        i2d_map = oidx_to_raveled_idx(r_idxs, s_idxs, shape)
        d2i_map = oidx_to_raveled_idx(np.argsort(r_idxs), s_idxs, shape)
        _samples = np.array(samples)
        s2i_map = HashTable(
            max=len(_samples) * 2,  # type: ignore | 2x size for perf > mem
            dtype=_samples.dtype,
        )
        s2i_map.add(_samples)
        return cls(
            full_region_idxs=r_idxs,
            full_sample_idxs=s_idxs,
            i2d_map=i2d_map,
            d2i_map=d2i_map,
            s2i_map=s2i_map,
        )

    @property
    def full_samples(self) -> NDArray[np.str_]:
        return self.s2i_map.keys

    @property
    def is_subset(self) -> bool:
        return (
            self.region_subset_idxs is not None or self.sample_subset_idxs is not None
        )

    @property
    def n_regions(self) -> int:
        if self.region_subset_idxs is None:
            return len(self.full_region_idxs)
        return len(self.region_subset_idxs)

    @property
    def n_samples(self) -> int:
        if self.sample_subset_idxs is None:
            return len(self.full_sample_idxs)
        return len(self.sample_subset_idxs)

    def region_idxs(self, mode: Literal["i2d", "d2i"]) -> NDArray[np.integer]:
        """Map from input region indices to on-disk region indices or vice versa."""
        if mode == "i2d":
            return np.unravel_index(self.i2d_map[:: self.n_samples], self.full_shape)[0]
        else:
            return np.unravel_index(self.d2i_map[:: self.n_samples], self.full_shape)[0]

    def sample_idxs(self, mode: Literal["i2d", "d2i"]) -> NDArray[np.integer]:
        """Map from input sample indices to on-disk sample indices or vice versa."""
        if mode == "i2d":
            return np.unravel_index(self.i2d_map[: self.n_samples], self.full_shape)[1]
        else:
            return np.unravel_index(self.d2i_map[: self.n_samples], self.full_shape)[1]

    @property
    def samples(self) -> List[str]:
        if self.sample_subset_idxs is None:
            return self.full_samples.tolist()
        return self.full_samples[self.sample_subset_idxs].tolist()

    @property
    def shape(self) -> tuple[int, int]:
        return self.n_regions, self.n_samples

    @property
    def full_shape(self) -> tuple[int, int]:
        return len(self.full_region_idxs), len(self.full_sample_idxs)

    def __len__(self):
        return len(self.i2d_map)

    def subset_to(
        self,
        regions: Optional[Idx] = None,
        samples: Optional[Idx] = None,
    ) -> "DatasetIndexer":
        """Subset the dataset to specific regions and/or samples."""
        if regions is None and samples is None:
            return self

        if samples is not None:
            sample_idxs = idx_like_to_array(samples, self.n_samples)
        else:
            sample_idxs = np.arange(self.n_samples, dtype=np.intp)

        if regions is not None:
            region_idxs = idx_like_to_array(regions, self.n_regions)
        else:
            region_idxs = np.arange(self.n_regions, dtype=np.intp)

        idx = oidx_to_raveled_idx(
            row_idx=region_idxs,
            col_idx=sample_idxs,
            shape=self.shape,
        )

        i2d_map = self.i2d_map[idx]

        return evolve(
            self,
            i2d_map=i2d_map,
            region_subset_idxs=region_idxs,
            sample_subset_idxs=sample_idxs,
        )

    def to_full_dataset(self) -> "DatasetIndexer":
        """Return a full sized dataset, undoing any subsettting."""
        i2d_map = oidx_to_raveled_idx(
            self.full_region_idxs, self.full_sample_idxs, self.full_shape
        )
        return evolve(
            self,
            i2d_map=i2d_map,
            region_subset_idxs=None,
            sample_subset_idxs=None,
        )

    def parse_idx(
        self, idx: Idx | tuple[Idx] | tuple[Idx, StrIdx]
    ) -> tuple[NDArray[np.integer], bool, tuple[int, ...] | None]:
        if not isinstance(idx, tuple):
            regions = idx
            samples = slice(None)
        elif len(idx) == 1:
            regions = idx[0]
            samples = slice(None)
        else:
            regions, samples = idx

        s_idx = s2i(samples, self.s2i_map)
        idx = self.i2d_map.reshape(self.shape)[regions, s_idx]

        out_reshape = None
        squeeze = False
        if idx.ndim > 1:
            out_reshape = idx.shape
        elif idx.ndim == 0:
            squeeze = True

        idx = idx.ravel()

        return idx, squeeze, out_reshape

    def s2i(self, samples: StrIdx) -> Idx:
        """Convert sample names to sample indices."""
        return s2i(samples, self.s2i_map)


def s2i(str_idx: StrIdx, map: HashTable) -> Idx:
    """Convert a string index to an integer index using a hirola.HashTable."""
    if (
        isinstance(str_idx, str)
        or (isinstance(str_idx, np.ndarray) and is_dtype(str_idx, np.str_))
        or (isinstance(str_idx, Sequence) and isinstance(next(collapse(str_idx)), str))
    ):
        idx = map.get(str_idx)
        if (np.atleast_1d(idx) == -1).any():
            raise KeyError(
                f"Some keys not found in mapping: {np.unique(np.array(str_idx)[idx == -1])}"
            )
    else:
        idx = str_idx

    idx = cast(Idx, idx)  # above clause does this, but can't narrow type

    return idx



import numpy as np
import pytest
from genvarloader._dataset._indexing import DatasetIndexer
from genvarloader._utils import idx_like_to_array
from pytest_cases import fixture, parametrize_with_cases


@fixture
def dsi():
    n_samples = 2
    r_idx = np.array([1, 2, 0])
    s_idx = np.arange(n_samples)
    samples = ["Aang", "Katara"]
    return DatasetIndexer.from_region_and_sample_idxs(r_idx, s_idx, samples)


def case_1_region_all_samples():
    r_idx = 0
    s_idx = None
    expected_idx_map = np.array([2, 3])
    return r_idx, s_idx, expected_idx_map


def case_1_region_1_sample():
    r_idx = 0
    s_idx = 1
    expected_idx_map = np.array([3])
    return r_idx, s_idx, expected_idx_map


def case_all_regions_1_sample():
    r_idx = None
    s_idx = 1
    expected_idx_map = np.array([3, 5, 1])
    return r_idx, s_idx, expected_idx_map


def case_2_regions_1_sample():
    r_idx = [0, 2]
    s_idx = 1
    expected_idx_map = np.array([3, 1])
    return r_idx, s_idx, expected_idx_map


def case_2_regions_all_samples():
    r_idx = [0, 2]
    s_idx = None
    expected_idx_map = np.array([2, 3, 0, 1])
    return r_idx, s_idx, expected_idx_map


@parametrize_with_cases("r_idx, s_idx, expected_idx_map", cases=".")
def test_subset(dsi: DatasetIndexer, r_idx, s_idx, expected_idx_map):
    subset = dsi.subset_to(r_idx, s_idx)
    if r_idx is not None:
        r_idx = idx_like_to_array(r_idx, dsi.n_regions)
    else:
        r_idx = np.arange(dsi.n_regions, dtype=np.intp)
    if s_idx is not None:
        s_idx = idx_like_to_array(s_idx, dsi.n_samples)
    else:
        s_idx = np.arange(dsi.n_samples, dtype=np.intp)
    assert subset.n_regions == len(r_idx)
    assert subset.n_samples == len(s_idx)
    np.testing.assert_equal(subset.i2d_map, expected_idx_map)


def test_subset_to_full(dsi: DatasetIndexer):
    subset = dsi.subset_to([0, 2], [0])
    full = subset.to_full_dataset()
    assert full.n_regions == dsi.n_regions
    assert full.n_samples == dsi.n_samples
    np.testing.assert_equal(full.i2d_map, dsi.i2d_map)


def getitem_squeeze():
    regions = 0
    samples = "Aang"
    s_idx = 0
    return regions, samples, s_idx


def getitem_1d():
    regions = 0
    samples = ["Aang", "Katara"]
    s_idx = [0, 1]
    return regions, samples, s_idx


def getitem_slice_scalar():
    regions = slice(2)
    samples = "Aang"
    s_idx = 0
    return regions, samples, s_idx


def getitem_reshape():
    regions = 0
    samples = [["Aang", "Katara"], ["Katara", "Aang"]]
    s_idx = [[0, 1], [1, 0]]
    return regions, samples, s_idx


@pytest.mark.xfail(strict=True, raises=KeyError, reason="Zuko is not in the dataset")
def getitem_missing():
    regions = 0
    samples = ["Aang", "Zuko"]
    s_idx = [0, -1]
    return regions, samples, s_idx


@parametrize_with_cases("regions, samples, s_idx", cases=".", prefix="getitem_")
def test_getitem_shape(dsi: DatasetIndexer, regions, samples, s_idx):
    idx, squeeze, reshape = dsi.parse_idx((regions, samples))
    desired = dsi.i2d_map.reshape(dsi.shape)[regions, s_idx]

    if reshape is not None:
        assert desired.shape == reshape
    elif squeeze:
        assert idx.shape == (1,)

    np.testing.assert_equal(idx, desired.ravel(), strict=True)



from enum import Enum
from pathlib import Path
from typing import Annotated, Optional

from typer import Argument, Option, Typer

__all__ = []


app = Typer(rich_markup_mode="rich")


class LOG_LEVEL(str, Enum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


@app.command(no_args_is_help=True)
def main(
    path: Annotated[
        Path,
        Argument(
            help="Path to the output directory, using the .gvl extension for clarity is encouraged."
        ),
    ],
    bed: Annotated[Path, Argument(help="Path to the BED file.")],
    variants: Annotated[
        Optional[Path], Option(help="Path to variants file, either VCF or PGEN.")
    ] = None,
    bigwig_table: Annotated[
        Optional[Path],
        Option(
            help='CSV or TSV with columns "sample" and "path" mapping each sample to its BigWig.'
        ),
    ] = None,
    samples: Annotated[
        Optional[str],
        Option(help="Comma-separated list of samples to include/subset to."),
    ] = None,
    length: Annotated[
        Optional[int],
        Option(
            help="Length of the sequences. If not provided, will default to maximum length seen in the BED file."
        ),
    ] = None,
    max_jitter: Annotated[
        int,
        Option(
            help="Maximum jitter to allow. Permitted by expanding the region length by 2 * max jitter."
        ),
    ] = 0,
    overwrite: Annotated[
        bool,
        Option(help="Overwrite the output directory if it exists."),
    ] = False,
    max_memory: Annotated[
        str,
        Option(
            help="Hint for maximum memory to use. Can be a number or use suffixes M and G to specify units. Actual usage will be marginally higher than this. Default is 4 GiB."
        ),
    ] = "4G",
    log_level: Annotated[
        LOG_LEVEL,
        Option(
            help="Log level to use. One of DEBUG, INFO, WARNING, ERROR, CRITICAL. Default is INFO."
        ),
    ] = LOG_LEVEL.INFO,
):
    """Write a GenVarLoader dataset from a BED3+ file and a VCF file and/or BigWig files.

    If a VCF is included, the dataset will support yielding haplotypes.
    If BigWigs are included, the dataset will support yielding base-pair resolution tracks.
    One of either a VCF or BigWigs must be provided.

    [b]Sample subsetting behavior:[/b]
    If a VCF and BigWigs are provided, the samples will be subset to the intersection of the samples in the VCF and BigWigs.
    If a list of specific samples are provided via --samples, that subset will take precedence.
    """
    if path.exists() and not overwrite:
        raise FileExistsError(
            f"Output directory {path} already exists. Use --overwrite to overwrite."
        )

    import sys

    from loguru import logger

    from genvarloader._bigwig import BigWigs
    from genvarloader._dataset._write import write

    if bigwig_table is not None:
        bigwigs = BigWigs.from_table("bws", bigwig_table)
    else:
        bigwigs = None

    if samples is not None:
        _samples = samples.split(",")
    else:
        _samples = None

    if max_memory[-1] == "M":
        _max_memory = int(max_memory[:-1]) * 1024**2
    elif max_memory[-1] == "G":
        _max_memory = int(max_memory[:-1]) * 1024**3
    else:
        _max_memory = int(max_memory)

    logger.remove()
    logger.add(sys.stderr, level=log_level.value)
    logger.enable("genvarloader")

    write(
        path=path,
        bed=bed,
        variants=variants,
        bigwigs=bigwigs,
        samples=_samples,
        length=length,
        max_jitter=max_jitter,
        overwrite=overwrite,
        max_mem=_max_memory,
    )


if __name__ == "__main__":
    app()



from pathlib import Path

import numpy as np
import polars as pl
import seqpro as sp
from einops import repeat
from natsort import natsorted

from ._dataset._genotypes import SparseGenotypes
from ._dataset._impl import RaggedDataset
from ._dataset._indexing import DatasetIndexer
from ._dataset._intervals import tracks_to_intervals
from ._dataset._reconstruct import Haps, HapsTracks, Reference, Tracks, _Variants
from ._dataset._utils import bed_to_regions
from ._ragged import Ragged, RaggedIntervals
from ._utils import _lengths_to_offsets
from ._variants._records import VLenAlleles


def get_dummy_dataset():
    """Return a dummy :class:`Dataset <genvarloader.Dataset>`  with 4 regions, 4 samples, max jitter of 2, a reference genome of all :code:`"N"`, genotypes, and
    1 track "read-depth" where each track is :code:`[1, 2, 3, 4, 5, 6]` in the reference coordinate system, where :code:`3` is aligned
    with each region's start coordinate. Is initialized to return ragged haplotypes and tracks with no jitter and deterministic reconstruction algorithms.
    """
    max_jitter = 2

    dummy_samples = ["Aang", "Katara", "Sokka", "Toph"]

    dummy_contigs = [str(i) for i in range(1, 23)] + ["X", "Y", "MT"]
    dummy_bed = pl.DataFrame(
        {
            "chrom": ["8", "3", "6", "2"],
            "chromStart": [5, 13, 8, 2],
            "chromEnd": [8, 16, 11, 5],
            "strand": ["+", "-", "+", "+"],
        }
    )

    with pl.StringCache():
        pl.Series(natsorted(dummy_contigs), dtype=pl.Categorical())
        sorted_bed = dummy_bed.with_row_index().sort(
            pl.col("chrom").cast(pl.Categorical())
        )

    r_idx_map = np.argsort(sorted_bed["index"])
    sorted_bed = sorted_bed.drop("index")
    dummy_idxer = DatasetIndexer.from_region_and_sample_idxs(
        r_idx_map, np.arange(len(dummy_samples)), dummy_samples
    )

    dummy_regions = bed_to_regions(sorted_bed, dummy_contigs)

    ref_len = 20
    ref_lens = np.full(len(dummy_contigs), ref_len, dtype=np.int32)
    ref = np.full(ref_len * len(dummy_contigs), b"N", dtype="S1").view(np.uint8)
    dummy_ref = Reference(
        reference=ref,
        contigs=dummy_contigs,
        offsets=_lengths_to_offsets(ref_lens, np.uint64),
        pad_char=ord(b"N"),
    )

    dummy_vars = _Variants(
        positions=repeat(dummy_regions[:, 1], "r -> (r s)", s=4),
        sizes=repeat(np.array([-2, -1, 0, 1], np.int32), "s -> (r s)", r=4),
        alts=VLenAlleles(
            alleles=repeat(sp.cast_seqs("ACGTT"), "a -> (r a)", r=4),
            offsets=_lengths_to_offsets(
                repeat(np.array([1, 1, 1, 2]), "s -> (r s)", r=4)
            ),
        ),
    )

    v_idxs = (
        np.array([[3, 2, 4, 1], [1, 3, 2, 4], [2, 1, 4, 3], [4, 2, 3, 1]])[
            [3, 1, 2, 0]
        ]  # target lengths
        - 1  # idx within region
        + 4 * np.arange(4)[:, None]  # adjust by region/contig offset
    )
    dummy_genos = SparseGenotypes(
        variant_idxs=v_idxs.ravel(),
        offsets=np.arange(0, 4 * 4 + 1, dtype=np.int64),  # every entry has 1 variant
        n_regions=4,
        n_samples=4,
        ploidy=1,
    )

    dummy_haps = Haps(dummy_ref, dummy_vars, dummy_genos, False)

    # (r s), want tracks of [1, 2, 3, 4, 5] for each region so that pad values of 0 are obvious
    track_regions = dummy_regions.copy()
    track_regions[:, 1] -= max_jitter
    track_regions[:, 2] = track_regions[:, 1] + 5 + max_jitter
    t_len = 5
    data, offsets = tracks_to_intervals(
        regions=track_regions,
        tracks=repeat(
            np.arange(1, 1 + t_len + 2 * max_jitter, dtype=np.float32),
            "l -> (r l)",
            r=4,
        ),
        track_offsets=_lengths_to_offsets(np.full(4, t_len + 2 * max_jitter)),
    )
    lengths = np.diff(offsets)
    data = repeat(data, "(r i) -> (r s i)", r=4, s=4)
    offsets = _lengths_to_offsets(repeat(lengths, "r -> (r s)", s=4))
    dummy_itvs = {
        "read-depth": RaggedIntervals.from_offsets(
            data=data, shape=(4, 4), offsets=offsets
        )
    }

    dummy_tracks = Tracks(dummy_itvs, ["read-depth"])

    dummy_recon = HapsTracks(dummy_haps, dummy_tracks)

    dummy_dataset: RaggedDataset[Ragged[np.bytes_], Ragged[np.float32], None, None] = (
        RaggedDataset(
            path=Path("dummy"),
            output_length="ragged",
            max_jitter=max_jitter,
            return_indices=False,
            contigs=dummy_contigs,
            jitter=0,
            deterministic=True,
            rc_neg=True,
            transform=None,
            _full_bed=dummy_bed,
            _full_regions=dummy_regions,
            _jittered_regions=dummy_regions.copy(),
            _idxer=dummy_idxer,
            _seqs=dummy_haps,
            _tracks=dummy_tracks,
            _recon=dummy_recon,
            _rng=np.random.default_rng(),
        )
    )

    return dummy_dataset



from pathlib import Path
from tempfile import NamedTemporaryFile

from genvarloader._variants._utils import path_is_pgen, path_is_vcf


def test_path_is_pgen():
    assert path_is_pgen(Path("test.pgen"))

    with NamedTemporaryFile(suffix=".pgen") as f:
        assert path_is_pgen(f.name[:-5])


def test_path_is_vcf():
    assert path_is_vcf(Path("test.vcf"))
    assert path_is_vcf(Path("test.vcf.gz"))
    assert path_is_vcf(Path("test.bcf"))
    assert path_is_vcf(Path("test.bcf.gz"))



from typing import Optional

import numba as nb
import numpy as np
from numpy.typing import NDArray

__all__ = []


@nb.njit(parallel=True, nogil=True, cache=True)
def shift_and_realign_tracks_sparse(
    out: NDArray[np.float32],
    out_offsets: NDArray[np.int64],
    regions: NDArray[np.int32],
    shifts: NDArray[np.int32],
    geno_offset_idxs: NDArray[np.integer],
    geno_v_idxs: NDArray[np.int32],
    geno_offsets: NDArray[np.int64],
    positions: NDArray[np.int32],
    sizes: NDArray[np.int32],
    tracks: NDArray[np.float32],
    track_offsets: NDArray[np.int64],
    keep: Optional[NDArray[np.bool_]] = None,
    keep_offsets: Optional[NDArray[np.int64]] = None,
):
    """Shift and realign tracks to correspond to haplotypes.

    Parameters
    ----------
    out : NDArray[np.float32]
        Ragged array with shape (batch, ploidy). Shifted and re-aligned tracks.
    out_offsets : NDArray[np.int64]
        Shape = (batch*ploidy + 1) Offsets into out.
    regions : NDArray[np.int32]
        Shape = (batch, 3) Regions, each is (contig_idx, start, end).
    shifts : NDArray[np.int32]
        Shape = (batch, ploidy) Shifts for each haplotype.
    geno_offset_idxs : NDArray[np.intp]
        Shape = (batch, ploidy) Indices into offsets for each region.
    geno_v_idxs : NDArray[np.int32]
        Shape = (variants) Indices of variants.
    geno_offsets : NDArray[np.uint32]
        Shape = (tot_regions*samples*ploidy + 1) Offsets into variant idxs.
    positions : NDArray[np.int32]
        Shape = (total_variants) Positions of variants.
    sizes : NDArray[np.int32]
        Shape = (total_variants) Sizes of variants.
    tracks : NDArray[np.float32]
        Shape = (batch*ploidy*length) Tracks.
    track_offsets : NDArray[np.int64]
        Shape = (batch + 1) Offsets into tracks.
    keep : Optional[NDArray[np.bool_]]
        Shape = (batch*ploidy*variants) Keep mask for genotypes.
    keep_offsets : Optional[NDArray[np.int64]]
        Shape = (batch*ploidy + 1) Offsets into keep.
    """
    n_regions, ploidy = geno_offset_idxs.shape
    for query in nb.prange(n_regions):
        t_s, t_e = track_offsets[query], track_offsets[query + 1]
        q_track = tracks[t_s:t_e]
        # assumes start is never altered upstream by differing hap lengths (true for left-aligned variants)
        q_start = regions[query, 1]

        for hap in nb.prange(ploidy):
            o_idx = geno_offset_idxs[query, hap]

            k_idx = query * ploidy + hap
            if keep is not None and keep_offsets is not None:
                qh_keep = keep[keep_offsets[k_idx] : keep_offsets[k_idx + 1]]
            else:
                qh_keep = None

            out_s, out_e = out_offsets[k_idx], out_offsets[k_idx + 1]
            qh_out = out[out_s:out_e]
            qh_shifts = shifts[query, hap]

            shift_and_realign_track_sparse(
                offset_idx=o_idx,
                geno_v_idxs=geno_v_idxs,
                geno_offsets=geno_offsets,
                positions=positions,
                sizes=sizes,
                shift=qh_shifts,
                track=q_track,
                query_start=q_start,
                out=qh_out,
                keep=qh_keep,
            )


@nb.njit(nogil=True, cache=True)
def shift_and_realign_track_sparse(
    offset_idx: int,
    geno_v_idxs: NDArray[np.int32],
    geno_offsets: NDArray[np.int64],
    positions: NDArray[np.int32],
    sizes: NDArray[np.int32],
    shift: int,
    track: NDArray[np.float32],
    query_start: int,
    out: NDArray[np.float32],
    keep: Optional[NDArray[np.bool_]] = None,
):
    """Shift and realign a track to correspond to a haplotype.

    Parameters
    ----------
    offset_idx : NDArray[np.int32]
        Shape = (n_variants) Genotypes of variants.
    positions : NDArray[np.int32]
        Shape = (total_variants) Positions of variants.
    sizes : NDArray[np.int32]
        Shape = (total_variants) Sizes of variants.
    shift : int
        Total amount to shift by.
    track : NDArray[np.float32]
        Shape = (length) Track.
    out : NDArray[np.uint8]
        Shape = (out_length) Shifted and re-aligned track.
    keep : Optional[NDArray[np.bool_]]
        Shape = (n_variants) Keep mask for genotypes.
    """
    _variant_idxs = geno_v_idxs[geno_offsets[offset_idx] : geno_offsets[offset_idx + 1]]
    length = len(out)
    n_variants = len(_variant_idxs)

    if n_variants == 0:
        # guaranteed to have shift = 0
        out[:] = track[:length]
        return

    # where to get next track value
    track_idx = 0
    # where to put next value
    out_idx = 0
    # how much we've shifted
    shifted = 0

    for v in range(n_variants):
        if keep is not None and not keep[v]:
            continue

        variant: np.int32 = _variant_idxs[v]

        # position of variant relative to ref from fetch(contig, start, q_end)
        # i.e. has been put into same coordinate system as ref_idx
        v_rel_pos = positions[variant] - query_start
        v_diff = sizes[variant]
        # +1 assumes atomized variants, exactly 1 nt shared between REF and ALT
        v_rel_end = v_rel_pos - min(0, v_diff) + 1

        # variant is a DEL spanning start
        if v_diff < 0 and v_rel_pos < 0 and v_rel_end >= 0:
            track_idx = v_rel_end
            continue

        # overlapping variants
        # v_rel_pos < ref_idx only if we see an ALT at a given position a second
        # time or more. We'll do what bcftools consensus does and only use the
        # first ALT variant we find.
        if v_rel_pos < track_idx:
            continue

        v_len = max(0, v_diff) + 1

        # handle shift
        if shifted < shift:
            ref_shift_dist = v_rel_pos - track_idx
            # need more than variant to finish shift
            if shifted + ref_shift_dist + v_len < shift:
                # skip the variant
                continue
            # can finish shift without using variant
            elif shifted + ref_shift_dist >= shift:
                track_idx += shift - shifted
                shifted = shift
                # can still use the variant and whatever ref is left between
                # ref_idx and the variant
            # ref + (some of) variant is enough to finish shift
            else:
                # how much left to shift - amount of ref we can use
                allele_start_idx = shift - shifted - ref_shift_dist
                shifted = shift
                #! without if statement, parallel=True can cause a SystemError!
                # * parallel jit cannot handle changes in array dimension.
                # * without this, allele can change from a 1D array to a 0D
                # * array.
                if allele_start_idx == v_len:
                    # consume track up to end of variant
                    track_idx = v_rel_end
                    continue
                # consume track up to start of variant
                track_idx = v_rel_pos
                # adjust variant length
                v_len -= allele_start_idx

        # SNPs (but not MNPs because we don't have ALT length, MNPs are not atomic)
        # skipped because for tracks they always match the reference
        if v_diff == 0:
            continue

        # add track values up to variant
        track_len = v_rel_pos - track_idx
        if out_idx + track_len >= length:
            # track will get written by final clause
            # handles case where extraneous variants downstream of the haplotype were provided
            break
        out[out_idx : out_idx + track_len] = track[track_idx : track_idx + track_len]
        out_idx += track_len

        # indels (substitutions are skipped above and then handled by above clause)
        writable_length = min(v_len, length - out_idx)
        out[out_idx : out_idx + writable_length] = track[v_rel_pos]
        out_idx += writable_length
        track_idx = v_rel_end

        if out_idx >= length:
            break

    if shifted < shift:
        # need to shift the rest of the track
        track_idx += shift - shifted
        track_idx = min(track_idx, len(track))
        shifted = shift

    # fill rest with track and pad with 0
    unfilled_length = length - out_idx
    if unfilled_length > 0:
        writable_ref = min(unfilled_length, len(track) - track_idx)
        out_end_idx = out_idx + writable_ref
        ref_end_idx = track_idx + writable_ref
        out[out_idx:out_end_idx] = track[track_idx:ref_end_idx]

        if out_end_idx < length:
            out[out_end_idx:] = 0



# %%

import numpy as np
from einops import repeat
from genvarloader._dataset._genotypes import (
    DenseGenotypes,
    SparseGenotypes,
    SparseSomaticGenotypes,
)
from pytest_cases import parametrize_with_cases


# %%
def case_snps():
    n_samples = 2
    ploidy = 2
    n_regions = 1
    genos = np.array([0, 1, -9], np.int8)
    n_alt = (genos == 1).sum()
    n_variants = len(genos)
    genos = repeat(genos, "v -> s p v", s=n_samples, p=ploidy)
    first_v_idxs = np.arange(n_regions, dtype=np.int32)
    offsets = np.array([0, n_variants], np.int64)

    sparse_v_idxs = (genos == 1).nonzero()[-1].astype(np.int32)
    sparse_offsets = np.empty(n_regions * ploidy * n_samples + 1, np.int64)
    sparse_offsets[0] = 0
    sparse_offsets[1:] = n_alt * np.arange(1, n_regions * ploidy * n_samples + 1)
    dense = DenseGenotypes(
        genos=genos, first_v_idxs=first_v_idxs, offsets=offsets, n_samples=n_samples
    )
    sparse = SparseGenotypes(
        variant_idxs=sparse_v_idxs,
        offsets=sparse_offsets,
        n_regions=n_regions,
        n_samples=n_samples,
        ploidy=ploidy,
    )

    sparse_v_idxs = (genos == 1).any(1).nonzero()[-1].astype(np.int32)
    sparse_offsets = np.empty(n_regions * n_samples + 1, np.int64)
    sparse_offsets[0] = 0
    sparse_offsets[1:] = n_alt * np.arange(1, n_regions * n_samples + 1)
    ccfs = np.array([0, 0.5, -9.0], np.float32)
    ccfs = repeat(ccfs, "v -> s v", s=n_samples)
    sparse_somatic = SparseSomaticGenotypes(
        variant_idxs=sparse_v_idxs,
        ccfs=ccfs[:, [1]].T.ravel(),
        offsets=sparse_offsets,
        n_regions=n_regions,
        n_samples=n_samples,
    )

    ilens = np.zeros(n_variants, np.int32)
    positions = np.arange(n_variants, dtype=np.int32)
    starts = np.arange(n_regions, dtype=np.int32)
    lengths = np.full(n_regions, 3, np.int32)
    max_ends = np.array([3], np.int32)

    return (
        dense,
        sparse,
        sparse_somatic,
        ilens,
        positions,
        starts,
        lengths,
        max_ends,
        ccfs,
    )


def case_indels():
    n_samples = 2
    ploidy = 2
    n_regions = 1
    genos = np.array([0, 1, -9, 1], np.int8)
    n_alt = (genos > 0).sum()
    n_variants = len(genos)
    genos = repeat(genos, "v -> s p v", s=n_samples, p=ploidy)
    first_v_idxs = np.arange(n_regions, dtype=np.int32)
    offsets = np.array([0, n_variants], np.int64)

    sparse_v_idxs = (genos > 0).nonzero()[-1].astype(np.int32)
    sparse_offsets = np.empty(n_regions * ploidy * n_samples + 1, np.int64)
    sparse_offsets[0] = 0
    sparse_offsets[1:] = n_alt * np.arange(1, n_regions * ploidy * n_samples + 1)
    dense = DenseGenotypes(
        genos=genos, first_v_idxs=first_v_idxs, offsets=offsets, n_samples=n_samples
    )
    sparse = SparseGenotypes(
        variant_idxs=sparse_v_idxs,
        offsets=sparse_offsets,
        n_regions=n_regions,
        n_samples=n_samples,
        ploidy=ploidy,
    )

    sparse_v_idxs = (genos == 1).any(1).nonzero()[-1].astype(np.int32)
    sparse_offsets = np.empty(n_regions * n_samples + 1, np.int64)
    sparse_offsets[0] = 0
    sparse_offsets[1:] = n_alt * np.arange(1, n_regions * n_samples + 1)
    ccfs = np.array([0, 0.5, -9.0, 0.4], np.float32)
    ccfs = repeat(ccfs, "v -> s v", s=n_samples)
    sparse_somatic = SparseSomaticGenotypes(
        variant_idxs=sparse_v_idxs,
        ccfs=ccfs[:, [1, 3]].T.ravel(),
        offsets=sparse_offsets,
        n_regions=n_regions,
        n_samples=n_samples,
    )

    ilens = np.array([0, -2, 0, 2])
    positions = np.arange(n_variants, dtype=np.int32)
    starts = np.arange(n_regions, dtype=np.int32)
    lengths = np.full(n_regions, 4, np.int32)
    max_ends = np.array([4], np.int32)

    return (
        dense,
        sparse,
        sparse_somatic,
        ilens,
        positions,
        starts,
        lengths,
        max_ends,
        ccfs,
    )


@parametrize_with_cases(
    "dense, sparse, sparse_somatic, ilens, positions, starts, length, max_ends, dosages",
    cases=".",
)
def test_from_dense(
    dense: DenseGenotypes,
    sparse: SparseGenotypes,
    sparse_somatic: SparseSomaticGenotypes,
    ilens,
    positions,
    starts,
    length,
    max_ends,
    dosages,
):
    desired = sparse
    actual = SparseGenotypes.from_dense(dense.genos, dense.first_v_idxs, dense.offsets)

    np.testing.assert_equal(actual.offsets, desired.offsets)
    np.testing.assert_equal(actual.variant_idxs, desired.variant_idxs)


@parametrize_with_cases(
    "dense, sparse, sparse_somatic, ilens, positions, starts, lengths, max_ends, dosages",
    cases=".",
)
def test_from_dense_with_length(
    dense: DenseGenotypes,
    sparse: SparseGenotypes,
    sparse_somatic: SparseSomaticGenotypes,
    ilens,
    positions,
    starts,
    lengths,
    max_ends,
    dosages,
):
    desired = sparse
    desired_max_ends = max_ends

    actual, max_ends = SparseGenotypes.from_dense_with_length(
        dense.genos,
        dense.first_v_idxs,
        dense.offsets,
        ilens,
        positions,
        starts,
        lengths,
    )

    np.testing.assert_equal(actual.offsets, desired.offsets)
    np.testing.assert_equal(actual.variant_idxs, desired.variant_idxs)
    np.testing.assert_equal(max_ends, desired_max_ends)


@parametrize_with_cases(
    "dense, sparse, sparse_somatic, ilens, positions, starts, lengths, max_ends, dosages",
    cases=".",
)
def test_somatic_from_dense(
    dense: DenseGenotypes,
    sparse: SparseGenotypes,
    sparse_somatic: SparseSomaticGenotypes,
    ilens,
    positions,
    starts,
    lengths,
    max_ends,
    dosages,
):
    desired = sparse_somatic
    actual = SparseSomaticGenotypes.from_dense(
        dense.genos, dense.first_v_idxs, dense.offsets, dosages
    )

    np.testing.assert_equal(actual.offsets, desired.offsets)
    np.testing.assert_equal(actual.variant_idxs, desired.variant_idxs)
    np.testing.assert_equal(actual.ccfs, desired.ccfs)


@parametrize_with_cases(
    "dense, sparse, sparse_somatic, ilens, positions, starts, lengths, max_ends, dosages",
    cases=".",
)
def test_somatic_from_dense_with_length(
    dense: DenseGenotypes,
    sparse: SparseGenotypes,
    sparse_somatic: SparseSomaticGenotypes,
    ilens,
    positions,
    starts,
    lengths,
    max_ends,
    dosages,
):
    desired = sparse_somatic
    desired_max_ends = max_ends

    actual, max_ends = SparseSomaticGenotypes.from_dense_with_length(
        dense.genos,
        dense.first_v_idxs,
        dense.offsets,
        ilens,
        positions,
        starts,
        lengths,
        dosages,
    )

    np.testing.assert_equal(actual.offsets, desired.offsets)
    np.testing.assert_equal(actual.variant_idxs, desired.variant_idxs)
    np.testing.assert_equal(max_ends, desired_max_ends)


# %%
# test_from_dense_with_length(*case_snps())
# test_from_dense_with_length(*case_indels())



import re
import subprocess
from pathlib import Path
from textwrap import dedent
from typing import Dict, List, Literal, Optional, Union, cast, overload

import numpy as np
import polars as pl
import pyarrow as pa
from attrs import define
from loguru import logger
from numpy.typing import ArrayLike, NDArray
from tqdm.auto import tqdm
from typing_extensions import Self, assert_never

from .._utils import _lengths_to_offsets, _normalize_contig_name, _offsets_to_lengths

try:
    import cyvcf2

    CYVCF2_INSTALLED = True
except ImportError:
    CYVCF2_INSTALLED = False

__all__ = []


@define
class VLenAlleles:
    """Variable length alleles.

    Create VLenAlleles from a polars Series of strings:
    >>> alleles = VLenAlleles.from_polars(pl.Series(["A", "AC", "G"]))

    Create VLenAlleles from offsets and alleles:
    >>> offsets = np.array([0, 1, 3, 4], np.uint64)
    >>> alleles = np.frombuffer(b"AACG", "|S1")
    >>> alleles = VLenAlleles(offsets, alleles)

    Get a single allele:
    >>> alleles[0]
    b'A'

    Get a slice of alleles:
    >>> alleles[1:]
    VLenAlleles(offsets=array([0, 2, 3]), alleles=array([b'AC', b'G'], dtype='|S1'))
    """

    offsets: NDArray[np.int64]
    alleles: NDArray[np.bytes_]

    @overload
    def __getitem__(self, idx: int) -> NDArray[np.bytes_]: ...

    @overload
    def __getitem__(self, idx: slice) -> "VLenAlleles": ...

    def __getitem__(self, idx: Union[int, np.integer, slice]):
        if isinstance(idx, (int, np.integer)):
            return self.get_idx(idx)
        elif isinstance(idx, slice):
            return self.get_slice(idx)

    def get_idx(self, idx: Union[int, np.integer]):
        if idx >= len(self) or idx < -len(self):
            raise IndexError("Index out of range.")
        if idx < 0:
            idx = len(self) + idx
        return self.alleles[self.offsets[idx] : self.offsets[idx + 1]]

    def get_slice(self, slc: slice):
        start: Optional[int]
        stop: Optional[int]
        start, stop = slc.start, slc.stop
        if start is None:
            start = 0
        elif start < 0:
            start += len(self)

        # handle empty result
        if start >= len(self) or (stop is not None and stop <= start):
            return VLenAlleles(np.empty(0, np.int64), np.empty(0, "|S1"))

        if stop is None:
            stop = len(self)
        elif stop < 0:
            stop += len(self)
        stop += 1
        new_offsets = self.offsets[start:stop].copy()
        _start, _stop = new_offsets[0], new_offsets[-1]
        new_alleles = self.alleles[_start:_stop]
        new_offsets -= self.offsets[start]
        return VLenAlleles(new_offsets, new_alleles)

    def __len__(self):
        """Number of alleles."""
        return len(self.offsets) - 1

    @property
    def nbytes(self):
        return self.offsets.nbytes + self.alleles.nbytes

    @classmethod
    def from_polars(cls, alleles: pl.Series):
        offsets = np.empty(len(alleles) + 1, np.int64)
        offsets[0] = 0
        offsets[1:] = alleles.str.len_bytes().cast(pl.Int64).cum_sum().to_numpy()
        flat_alleles = np.frombuffer(
            alleles.str.concat("").to_numpy()[0].encode(), "S1"
        )
        return cls(offsets, flat_alleles)

    def to_polars(self):
        n_alleles = len(self)
        offset_buffer = pa.py_buffer(self.offsets)
        allele_buffer = pa.py_buffer(self.alleles)
        string_arr = pa.LargeStringArray.from_buffers(
            n_alleles, offset_buffer, allele_buffer
        )
        return pl.Series(string_arr)

    @staticmethod
    def concat(*vlen_alleles: "VLenAlleles"):
        if len(vlen_alleles) == 0:
            return VLenAlleles(np.array([0], np.int64), np.array([], "|S1"))
        elif len(vlen_alleles) == 1:
            return vlen_alleles[0]

        nuc_per_allele = np.concatenate(
            [_offsets_to_lengths(v.offsets) for v in vlen_alleles]
        )
        offsets = _lengths_to_offsets(nuc_per_allele, np.int64)
        alleles = np.concatenate([v.alleles for v in vlen_alleles])
        return VLenAlleles(offsets, alleles)


@define
class RecordInfo:
    """Information about sets of records corresponding to range queries.

    Attributes
    ----------
    positions : NDArray[np.int32]
        Shape: (n_variants)
        Positions of variants.
    size_diffs : NDArray[np.int32]
        Shape: (n_variants)
        Difference in length between ref and alt.
    refs : VLenAlleles
        Effective shape: (n_variants)
        Reference alleles.
    alts : VLenAlleles
        Effective shape: (n_variants)
        Alternate alleles.
    start_idxs : NDArray[np.int32]
        Shape: (n_queries)
        Start indices of the records for each query.
    end_idxs : NDArray[np.int32]
        Shape: (n_queries)
        End indices of the records for each query.
    offsets : NDArray[np.int32]
        Shape: (n_queries + 1).
        Offsets for the records such that records[offsets[i]:offsets[i+1]] are the records
        for the i-th query.
    """

    positions: NDArray[np.int32]  # (n_variants)
    size_diffs: NDArray[np.int32]  # (n_variants)
    refs: VLenAlleles
    alts: VLenAlleles
    start_idxs: NDArray[np.int32]  # (n_queries)
    end_idxs: NDArray[np.int32]  # (n_queries)
    offsets: NDArray[np.int64]  # (n_queries + 1)

    @property
    def n_queries(self):
        return len(self.start_idxs)

    @classmethod
    def empty(cls, n_queries: int):
        return cls(
            positions=np.empty(n_queries, np.int32),
            size_diffs=np.empty(n_queries, np.int32),
            refs=VLenAlleles(
                np.zeros(n_queries + 1, np.int64), np.empty(n_queries, "S1")
            ),
            alts=VLenAlleles(
                np.zeros(n_queries + 1, np.int64), np.empty(n_queries, "S1")
            ),
            start_idxs=np.empty(n_queries, np.int32),
            end_idxs=np.empty(n_queries, np.int32),
            offsets=np.zeros(n_queries + 1, np.int64),
        )

    @staticmethod
    def concat(
        *record_infos: "RecordInfo", how: Literal["separate", "merge"] = "separate"
    ):
        if len(record_infos) == 0:
            return RecordInfo.empty(0)
        elif len(record_infos) == 1:
            return record_infos[0]

        positions = np.concatenate([r.positions for r in record_infos])
        size_diffs = np.concatenate([r.size_diffs for r in record_infos])
        refs = VLenAlleles.concat(*(r.refs for r in record_infos))
        alts = VLenAlleles.concat(*(r.alts for r in record_infos))
        start_idxs = np.concatenate([r.start_idxs for r in record_infos])
        end_idxs = np.concatenate([r.end_idxs for r in record_infos])
        v_per_query = np.concatenate(
            [_offsets_to_lengths(r.offsets) for r in record_infos]
        )
        if how == "separate":
            offsets = _lengths_to_offsets(v_per_query)
        elif how == "merge":
            offsets = np.array([0, v_per_query.sum()], np.int64)
        else:
            assert_never(how)
        return RecordInfo(
            positions=positions,
            size_diffs=size_diffs,
            refs=refs,
            alts=alts,
            start_idxs=start_idxs,
            end_idxs=end_idxs,
            offsets=offsets,
        )


@define
class Records:
    contigs: List[str]
    contig_offsets: Dict[str, int]
    """Offsets for each contig corresponding to the underlying file(s). Thus, if a single file backs
    a Records instance, then the contig offsets will be non-zero and correspond to the first variant
    that is on that contig. If contig-specific files back the Records instance, then the contig offsets
    will all be zero."""

    v_starts: Dict[str, NDArray[np.int32]]
    """0-based positions of variants, sorted by start."""
    v_diffs: Dict[str, NDArray[np.int32]]
    """Difference in length |REF| - |ALT|, sorted by start"""
    # no multi-allelics
    ref: Dict[str, VLenAlleles]
    alt: Dict[str, VLenAlleles]
    ######################

    ## sorted by ends ##
    v_ends: Dict[str, NDArray[np.int32]]
    """0-based, end *inclusive* ends of variants, sorted by end."""
    e2s_idx: Dict[str, NDArray[np.int32]]
    """End to start index (sorted by starts). After searching for overlapping variants with Records.v_ends,
    this maps the found indices to indices of variants sorted by starts."""
    v_diffs_sorted_by_ends: Dict[str, NDArray[np.int32]]
    """Difference in length |REF| - |ALT|, sorted by end."""
    max_del_q: Dict[str, NDArray[np.intp]]
    """Pending deprecation. Maximum possible deletion length for a query including variants from index 0 to i."""
    ####################

    @property
    def n_variants(self) -> int:
        return sum(len(v) for v in self.v_starts.values())

    @classmethod
    def from_vcf(cls, vcf: Union[str, Path, Dict[str, Path]]) -> Self:
        if isinstance(vcf, (str, Path)):
            vcf = {"_all": Path(vcf)}

        if "_all" in vcf:
            multi_contig_source = True
        else:
            multi_contig_source = False

        vcf_suffix = re.compile(r"\.[vb]cf(\.gz)?$")
        arrow_paths = {
            c: p.parent / vcf_suffix.sub(".gvl.arrow", p.name) for c, p in vcf.items()
        }

        if cls.gvl_arrow_exists(vcf, arrow_paths):
            return cls.from_gvl_arrow(arrow_paths)

        if multi_contig_source:
            start_df = cls.read_vcf(vcf["_all"])
            start_dfs = {
                cast(str, c[0]): v
                for c, v in start_df.partition_by(
                    "#CHROM", as_dict=True, maintain_order=True
                ).items()
            }
            del start_df
        else:
            start_dfs: Dict[str, pl.DataFrame] = {}
            for contig, path in vcf.items():
                start_dfs[contig] = cls.read_vcf(path)

        start_dfs, end_dfs = cls.process_start_df(start_dfs)

        if multi_contig_source:
            path = vcf["_all"]
            pl.concat(start_dfs.values()).write_ipc(
                path.parent / vcf_suffix.sub(".gvl.arrow", path.name)
            )
            pl.concat(end_dfs.values()).write_ipc(
                path.parent / vcf_suffix.sub(".gvl.ends.arrow", path.name)
            )
        else:
            for s_df, e_df, path in zip(
                start_dfs.values(), end_dfs.values(), vcf.values()
            ):
                s_df.write_ipc(path.parent / vcf_suffix.sub(".gvl.arrow", path.name))
                e_df.write_ipc(
                    path.parent / vcf_suffix.sub(".gvl.ends.arrow", path.name)
                )

        return cls.from_var_df(start_dfs, end_dfs)

    @staticmethod
    def read_vcf(vcf_path: Path):
        if not CYVCF2_INSTALLED:
            raise ImportError(
                "cyvcf2 is not installed. Please install it with `pip install cyvcf2`"
            )

        vcf = cyvcf2.VCF(str(vcf_path))  # pyright: ignore
        try:
            n_variants = cast(int, vcf.num_records)
        except ValueError as no_idx_err:
            logger.info(f"VCF file {vcf_path} appears to be unindexed, indexing...")
            try:
                subprocess.run(
                    ["bcftools", "index", vcf_path], check=True, capture_output=True
                )
            except subprocess.CalledProcessError as bcftools_err:
                logger.error(f"Failed to index VCF file {vcf_path}.")
                logger.error(f"stdout:\n{bcftools_err.stdout.decode()}")
                logger.error(f"stderr:\n{bcftools_err.stderr.decode()}")
                raise bcftools_err from no_idx_err
            vcf = cyvcf2.VCF(str(vcf_path))  # pyright: ignore
            n_variants = cast(int, vcf.num_records)
        chroms: List[Optional[str]] = [None] * n_variants
        positions = np.empty(n_variants, dtype=np.int32)
        refs: List[Optional[str]] = [None] * n_variants
        alts: List[Optional[str]] = [None] * n_variants
        sv_or_unknown = False
        with tqdm(
            total=n_variants, desc=f"Scanning variants from {vcf_path.name}"
        ) as pbar:
            for i, v in enumerate(vcf):
                chroms[i] = cast(str, v.CHROM)
                positions[i] = cast(int, v.POS)  # 1-indexed
                if v.is_sv or v.var_type == "unknown":
                    if not sv_or_unknown:
                        logger.warning(
                            f"VCF file {vcf_path} contains structural or unknown variants. These variants will be ignored."
                        )
                        sv_or_unknown = True
                    # use placeholder that makes ilen = 0 so it doesn't affect anything downstream
                    alt = "N" * len(v.REF)
                else:
                    alt = cast(List[str], v.ALT)
                refs[i] = cast(str, v.REF)
                # TODO: multi-allelics. Also, missing ALT (aka the * allele)?
                if len(alt) != 1:
                    raise RuntimeError(
                        f"""VCF file {vcf_path} contains multi-allelic or overlappings
                        variants which are not yet supported by GenVarLoader. Normalize 
                        the VCF with `bcftools norm -f <reference.fa>
                        -a --atom-overlaps . -m - <file.vcf>`"""
                    )
                alts[i] = alt[0]
                pbar.update()
        return pl.DataFrame(
            {
                "#CHROM": chroms,
                "POS": positions,
                "REF": refs,
                "ALT": alts,
            }
        )

    @classmethod
    def from_pvar(cls, pvar: Union[str, Path, Dict[str, Path]]) -> Self:
        if isinstance(pvar, (str, Path)):
            pvar = {"_all": Path(pvar)}

        if "_all" in pvar:
            multi_contig_source = True
        else:
            multi_contig_source = False

        arrow_paths = {c: p.with_suffix(".gvl.arrow") for c, p in pvar.items()}

        if cls.gvl_arrow_exists(pvar, arrow_paths):
            return cls.from_gvl_arrow(arrow_paths)

        if multi_contig_source:
            start_df = cls.read_pvar(pvar["_all"])
            start_dfs = {
                cast(str, c[0]): v
                for c, v in start_df.partition_by(
                    "#CHROM", as_dict=True, maintain_order=True
                ).items()
            }
        else:
            start_dfs: Dict[str, pl.DataFrame] = {}
            for contig, path in pvar.items():
                start_dfs[contig] = cls.read_pvar(path)

        start_dfs, end_dfs = cls.process_start_df(start_dfs)

        if multi_contig_source:
            path = pvar["_all"]
            pl.concat(start_dfs.values()).write_ipc(path.with_suffix(".gvl.arrow"))
            pl.concat(end_dfs.values()).write_ipc(path.with_suffix(".gvl.ends.arrow"))
        else:
            for s_df, e_df, path in zip(
                start_dfs.values(), end_dfs.values(), pvar.values()
            ):
                s_df.write_ipc(path.with_suffix(".gvl.arrow"))
                e_df.write_ipc(path.with_suffix(".gvl.ends.arrow"))

        return cls.from_var_df(start_dfs, end_dfs)

    @staticmethod
    def read_pvar(pvar_path: Path) -> pl.DataFrame:
        with open(pvar_path, "r") as f:
            skip_rows = 0
            while f.readline().startswith("##"):
                skip_rows += 1

        logger.info("Reading .pvar file...")
        pvar = pl.read_csv(
            pvar_path,
            separator="\t",
            skip_rows=skip_rows,
            columns=["#CHROM", "POS", "REF", "ALT"],
            schema_overrides={"#CHROM": pl.Utf8, "POS": pl.Int32},
        )
        logger.info("Finished reading .pvar file.")
        if (pvar["ALT"].str.contains(",")).any():
            raise RuntimeError(
                f"""PGEN file {pvar_path} contains multi-allelic variants which are 
                not yet supported by GenVarLoader. Split your multi-allelic 
                variants with `bcftools norm -f <reference.fa> -a
                --atom-overlaps . -m - <file.vcf>` then remake the PGEN file
                with the `--vcf-half-call r` option."""
            )
        return pvar

    @staticmethod
    def gvl_arrow_exists(
        sources: Union[Path, Dict[str, Path]], arrow: Union[Path, Dict[str, Path]]
    ) -> bool:
        # TODO: check if files were created after the last breaking change to the gvl.arrow format
        # check if the files exist and are more recent than the sources
        if isinstance(arrow, Path):
            assert isinstance(sources, Path)
            if not arrow.exists():
                return False
            if not arrow.stat().st_mtime >= sources.stat().st_mtime:
                return False
        else:
            assert isinstance(sources, dict)
            for contig, p in arrow.items():
                if not p.exists():
                    return False
                if not p.stat().st_mtime > sources[contig].stat().st_mtime:
                    return False

        return True

    @classmethod
    def from_gvl_arrow(cls, arrow: Union[str, Path, Dict[str, Path]]) -> Self:
        if isinstance(arrow, (str, Path)):
            arrow = {"_all": Path(arrow)}

        if "_all" in arrow:
            multi_contig_source = True
        else:
            multi_contig_source = False

        arrow_paths: Dict[str, Path] = {}
        for c, p in arrow.items():
            if p.suffix == ".gvl":
                arrow_paths[c] = p.with_suffix(".gvl.arrow")
            elif p.suffixes[-2:] == [".gvl", ".arrow"]:
                arrow_paths[c] = p
            else:
                raise ValueError(
                    f"Arrow file {p} does not have the .gvl suffix. Please provide the correct path."
                )

        if multi_contig_source:
            path = arrow_paths["_all"]
            start_dfs = {
                cast(str, c[0]): v
                for c, v in pl.read_ipc(path)
                .partition_by("#CHROM", as_dict=True)
                .items()
            }
            end_dfs = {
                cast(str, c[0]): v
                for c, v in pl.read_ipc(
                    path.parent / path.name.replace(".gvl.arrow", ".gvl.ends.arrow")
                )
                .partition_by("#CHROM", as_dict=True)
                .items()
            }
        else:
            start_dfs: Dict[str, pl.DataFrame] = {}
            end_dfs: Dict[str, pl.DataFrame] = {}
            for contig, path in arrow_paths.items():
                start_dfs[contig] = pl.read_ipc(path)
                end_dfs[contig] = pl.read_ipc(
                    path.parent / path.name.replace(".gvl.arrow", ".gvl.ends.arrow")
                )

        return cls.from_var_df(start_dfs, end_dfs)

    @staticmethod
    def process_start_df(start_dfs: Dict[str, pl.DataFrame]):
        """_summary_

        Parameters
        ----------
        start_dfs : Dict[str, pl.DataFrame]
            Each dataframe should have columns: POS, REF, ALT

        Returns
        -------
        start_dfs : Dict[str, pl.DataFrame]
        end_dfs : Dict[str, pl.DataFrame]
        """
        end_dfs: Dict[str, pl.DataFrame] = {}
        for contig, df in start_dfs.items():
            df = df.with_columns(
                POS=pl.col("POS") - 1,  #! change to 0-indexed
                ILEN=(
                    pl.col("ALT").str.len_bytes().cast(pl.Int32)
                    - pl.col("REF").str.len_bytes().cast(pl.Int32)
                ),
            )
            start_dfs[contig] = df

            ends = (
                df.select(
                    "#CHROM",
                    "ILEN",
                    END=pl.col("POS")
                    - pl.col("ILEN").clip(upper_bound=0),  #! end-inclusive
                )
                .with_row_index("VAR_IDX")
                .select(
                    pl.all().sort_by("END"),
                    # make E2S_IDX relative to each contig
                    pl.int_range(0, pl.len(), dtype=pl.Int32)
                    .sort_by("END")
                    .reverse()
                    .rolling_min(df.height, min_samples=1)
                    .reverse()
                    .alias("E2S_IDX"),
                )
                .select("#CHROM", "END", "ILEN", "VAR_IDX", "E2S_IDX")
                .sort("END")
                .join(df.select("POS").with_row_index("VAR_IDX"), on="VAR_IDX")
            )

            _starts = ends["POS"].to_numpy()
            _ends = ends["END"].to_numpy()
            was_ends = np.empty(len(_ends) + 1, dtype=_ends.dtype)
            was_ends[0] = 0
            #! convert to end-exclusive, + 1
            was_ends[1:] = _ends + 1
            md_q = np.searchsorted(was_ends, _starts, side="right") - 1
            ends = ends.with_columns(MD_Q=md_q).select(  # type: ignore
                "#CHROM", "END", "MD_Q", "ILEN", "E2S_IDX"
            )

            end_dfs[contig] = ends

        return start_dfs, end_dfs

    @classmethod
    def from_var_df(
        cls, start_dfs: Dict[str, pl.DataFrame], end_dfs: Dict[str, pl.DataFrame]
    ) -> Self:
        ## sorted by starts ##
        v_starts: Dict[str, NDArray[np.int32]] = {}
        # difference in length between ref and alt, sorted by start
        v_diffs: Dict[str, NDArray[np.int32]] = {}
        # no multi-allelics
        ref: Dict[str, VLenAlleles] = {}
        alt: Dict[str, VLenAlleles] = {}
        ######################

        ## sorted by ends ##
        v_ends: Dict[str, NDArray[np.int32]] = {}
        v_diffs_sorted_by_ends: Dict[str, NDArray[np.int32]] = {}
        e2s_idx: Dict[str, NDArray[np.int32]] = {}
        max_del_q: Dict[str, NDArray[np.int32]] = {}
        ####################

        contig_offset = 0
        contig_offsets: Dict[str, int] = {}
        for contig, s_df in start_dfs.items():
            e_df = end_dfs[contig]
            contig_offsets[contig] = contig_offset
            contig_offset += s_df.height
            v_starts[contig] = s_df["POS"].to_numpy()
            v_diffs[contig] = s_df["ILEN"].to_numpy()
            v_ends[contig] = e_df["END"].to_numpy()
            v_diffs_sorted_by_ends[contig] = e_df["ILEN"].to_numpy()
            e2s_idx[contig] = np.empty(e_df.height + 1, dtype=np.int32)
            e2s_idx[contig][:-1] = e_df["E2S_IDX"].to_numpy()
            e2s_idx[contig][-1] = e_df.height
            max_del_q[contig] = e_df["MD_Q"].to_numpy()

            # no multi-allelics
            ref[contig] = VLenAlleles.from_polars(s_df["REF"])
            alt[contig] = VLenAlleles.from_polars(s_df["ALT"])

        return cls(
            contigs=list(start_dfs.keys()),
            contig_offsets=contig_offsets,
            v_starts=v_starts,
            v_diffs=v_diffs,
            ref=ref,
            alt=alt,
            v_ends=v_ends,
            v_diffs_sorted_by_ends=v_diffs_sorted_by_ends,
            e2s_idx=e2s_idx,
            max_del_q=max_del_q,
        )

    def vars_in_range(
        self,
        contig: str,
        starts: ArrayLike,
        ends: ArrayLike,
    ) -> tuple[NDArray[np.int32], NDArray[np.int32]] | None:
        starts = np.atleast_1d(np.asarray(starts, dtype=np.int32))
        ends = np.atleast_1d(np.asarray(ends, dtype=np.int32))

        _contig = _normalize_contig_name(contig, self.contigs)
        if _contig is None:
            return

        rel_s_idxs = self.find_relative_start_idx(_contig, starts)
        rel_e_idxs = self.find_relative_end_idx(_contig, ends)

        if rel_s_idxs.min() == rel_e_idxs.max():
            return

        # make idxs absolute
        s_idxs = rel_s_idxs + self.contig_offsets[_contig]
        e_idxs = rel_e_idxs + self.contig_offsets[_contig]

        return s_idxs, e_idxs

    def find_relative_start_idx(
        self, contig: str, starts: NDArray[np.int32]
    ) -> NDArray[np.int32]:
        s_idx_by_end = np.searchsorted(self.v_ends[contig], starts)
        rel_s_idx = self.e2s_idx[contig][s_idx_by_end]
        return rel_s_idx

    def find_relative_end_idx(
        self, contig: str, ends: NDArray[np.int32]
    ) -> NDArray[np.int32]:
        rel_e_idx = np.searchsorted(self.v_starts[contig], ends)
        return rel_e_idx

    def __repr__(self):
        return dedent(
            f"""
            {self.__class__.__name__}
            contigs: {self.contigs}
            n_variants: {self.n_variants}
            """
        ).strip()



from __future__ import annotations

from pathlib import Path
from typing import (
    Callable,
    Dict,
    Mapping,
    Protocol,
    Sequence,
    TypeVar,
    Union,
)

import numpy as np
from attrs import define
from numpy.typing import ArrayLike, DTypeLike, NDArray

__all__ = ["Reader"]


DTYPE = TypeVar("DTYPE", bound=np.generic)
Idx = Union[
    int, np.integer, Sequence[int], slice, NDArray[np.integer], NDArray[np.bool_]
]
StrIdx = Idx | str | Sequence[str]
ListIdx = Union[Sequence[int], NDArray[np.integer]]


@define
class AnnotatedHaps:
    haps: NDArray[np.bytes_]
    var_idxs: NDArray[np.int32]
    ref_coords: NDArray[np.int32]

    @property
    def shape(self):
        return self.haps.shape

    def reshape(self, *shape: int | tuple[int, ...]):
        return AnnotatedHaps(
            self.haps.reshape(*shape),
            self.var_idxs.reshape(*shape),
            self.ref_coords.reshape(*shape),
        )

    def squeeze(self, axis: int | tuple[int, ...] | None = None) -> AnnotatedHaps:
        return AnnotatedHaps(
            self.haps.squeeze(axis),
            self.var_idxs.squeeze(axis),
            self.ref_coords.squeeze(axis),
        )


class Reader(Protocol):
    """Implements the read() method for returning data aligned to genomic coordinates."""

    name: str
    """Name of the reader, corresponding to the name of the DataArrays it returns."""
    dtype: DTypeLike
    """Data type of what the reader returns."""
    contigs: Mapping[str, int]
    sizes: Dict[str, int]
    """Sizes of the dimensions/axes of what the reader returns."""
    coords: Dict[str, NDArray]
    """Coordinates of what the reader returns, i.e. dimension labels."""
    rev_strand_fn: Callable[[NDArray], NDArray]
    """Function to reverse (and potentially complement) data for a genomic region. This
        is used when the strand is negative."""
    chunked: bool
    """Whether the reader acts like a chunked array store, in which sequential reads
        are far more performant than random access."""

    def read(
        self,
        contig: str,
        starts: ArrayLike,
        ends: ArrayLike,
        **kwargs,
    ) -> NDArray:
        """Read data corresponding to given genomic coordinates, akin to orthogonal indexing.
        The output shape will have length as the final dimension/axis i.e. (..., length).

        Parameters
        ----------
        contig : str
            Name of the contig/chromosome.
        starts : ArrayLike
            Start coordinates, 0-based.
        ends : ArrayLike
            End coordinates, 0-based exclusive.
        **kwargs
            Additional keyword arguments. For example, which samples or ploid numbers to
            return.

        Returns
        -------
        NDArray
            Data corresponding to the given genomic coordinates. The final axis is the
            length axis i.e. has length == (ends - starts).sum().

        Notes
        -----
        When multiple regions are provided (i.e. multiple starts and ends) they should
        be concatenated together in the output array along the length dimension.
        """
        ...


class _ToZarr(Protocol):
    """Implements the to_zarr() method."""

    def to_zarr(self, store: Path):
        """Materialize genomes-wide data as a Zarr store.

        Parameters
        ----------
        store : Path
            Directory to write the Zarr store.
        """
        ...



import random

import numba as nb
import numpy as np
from genvarloader._dataset._genotypes import _choose_unphased_variants
from numpy.typing import NDArray
from pytest_cases import parametrize_with_cases


def case_snps():
    variant_idxs = np.array([0, 1, 2], np.int32)
    ccfs = np.array([0.3, 0.2, 0.1], np.float32)
    positions = np.array([1, 2, 3], np.int32)
    sizes = np.array([0, 0, 0], np.int32)
    query_start = 0
    query_end = 4

    desired = np.ones(3, np.bool_)

    return variant_idxs, ccfs, positions, sizes, query_start, query_end, desired


def case_extra_snps():
    variant_idxs = np.array([0, 1, 2], np.int32)
    ccfs = np.array([0.3, 0.2, 0.1], np.float32)
    positions = np.array([1, 2, 3], np.int32)
    sizes = np.array([0, 0, 0], np.int32)
    query_start = 0
    query_end = 2

    desired = np.array([True, False, False], np.bool_)

    return variant_idxs, ccfs, positions, sizes, query_start, query_end, desired


def case_indels_0():
    # group 0
    # 0 1 2 3 4  5
    # r d - - - r/s

    # group 1
    # 0 1 2 3  4  5
    # r r r ii r r/s

    variant_idxs = np.array([0, 1, 2], np.int32)
    ccfs = np.array([0.3, 0.2, 0.1], np.float32)
    positions = np.array([1, 3, 5], np.int32)
    sizes = np.array([-3, 1, 0], np.int32)
    query_start = 0
    query_end = 8

    # v2 group assignment = 1 rand
    # group selection = 1 rand
    _seed(0)
    rands = _random(2)
    if rands[0] < 0.5:
        groups = np.array([0, 1, 1], np.uint32)
        ref_ends = np.array([5, 6, 0], np.uint32)
    else:
        groups = np.array([0, 1, 0], np.uint32)
        ref_ends = np.array([6, 4, 0], np.uint32)
    n_groups = 2

    cum_prop = np.empty(n_groups, np.float32)
    for g in range(n_groups):
        v_starts = positions[variant_idxs[groups == g]]
        v_ends = v_starts - np.minimum(0, sizes[variant_idxs[groups == g]]) + 1
        ref_lengths = np.minimum(v_ends, ref_ends[g]) - np.maximum(
            v_starts, query_start
        )
        cum_prop[g] = (ccfs[groups == g] / ref_lengths).sum()
    cum_prop = cum_prop.cumsum()
    cum_prop /= cum_prop[-1]
    keep_group = (rands[1] <= cum_prop).sum() - 1
    desired = groups == keep_group

    return variant_idxs, ccfs, positions, sizes, query_start, query_end, desired


def case_del_span_start():
    # group 0
    # 0 1 2 3 4 ...
    # . . - r s ...

    variant_idxs = np.array([0, 1, 2], np.int32)
    ccfs = np.array([0.3, 0.2, 0.1], np.float32)
    positions = np.array([0, 1, 4], np.int32)
    sizes = np.array([-2, 0, 0], np.int32)
    query_start = 2
    query_end = 12

    desired = np.array([True, False, True], np.bool_)

    return variant_idxs, ccfs, positions, sizes, query_start, query_end, desired


@parametrize_with_cases(
    "variant_idxs, ccfs, positions, sizes, query_start, query_end, desired",
    cases=".",
)
def test_mark_keep_variants(
    variant_idxs: NDArray[np.int32],
    ccfs: NDArray[np.float32],
    positions: NDArray[np.int32],
    sizes: NDArray[np.int32],
    query_start: int,
    query_end: int,
    desired: NDArray[np.bool_],
):
    n_variants = len(variant_idxs)

    groups = np.empty(n_variants, np.uint32)
    ends = np.empty(n_variants, np.uint32)
    write_lens = np.empty(n_variants, np.uint32)

    _seed(0)
    actual = _choose_unphased_variants(
        variant_idxs=variant_idxs,
        ccfs=ccfs,
        positions=positions,
        sizes=sizes,
        query_start=query_start,
        groups=groups,
        ref_ends=ends,
        write_lens=write_lens,
        query_end=query_end,
        deterministic=False,
    )

    np.testing.assert_array_equal(actual, desired)


@nb.njit
def _seed(seed: int):
    random.seed(seed)


@nb.njit
def _random(n_values: int):
    return [random.random() for _ in range(n_values)]



import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Protocol, Tuple, Union

import cyvcf2
import joblib
import numpy as np
import pgenlib
from numpy.typing import ArrayLike, NDArray

from .._utils import _get_rel_starts, _lengths_to_offsets


class CCFFieldError(Exception):
    """Exception raised when the CCF field is not found in the VCF record."""


__all__ = []


class Genotypes(Protocol):
    chunked: bool
    samples: NDArray[np.str_]
    ploidy: int
    handles: Optional[Dict[str, Any]]

    @property
    def n_samples(self) -> int:
        return len(self.samples)

    def close(self):
        """Close any open file handles and set handles to None. Should be a no-op if
        the handles are already closed or the interface does not use handles."""
        ...

    def read(
        self,
        contig: str,
        start_idxs: ArrayLike,
        end_idxs: ArrayLike,
        sample_idx: Optional[ArrayLike] = None,
        haplotype_idx: Optional[ArrayLike] = None,
    ) -> NDArray[np.int8]:
        """Read genotypes from a contig from index i to j, 0-based exclusive.

        Parameters
        ----------
        contig : str
            Name of the contig/chromosome.
        start_idxs : NDArray[np.intp]
            Start indices, 0-based.
        end_idxs : NDArray[np.intp]
            End indices, 0-based exclusive.
        sample_idx : NDArray[np.intp], optional
            Indices of the samples to include. Must be unique.
        haplotype_idx : NDArray[np.intp], optional
            Indices of the haplotypes to include. Must be unique.

        Returns
        -------
        genotypes : NDArray[np.int8]
            Shape: (samples ploidy variants). Genotypes for each query region.
        """
        ...

    def read_genos_and_dosages(
        self,
        contig: str,
        start_idxs: ArrayLike,
        end_idxs: ArrayLike,
        dosage_field: str,
        sample_idx: Optional[ArrayLike] = None,
        haplotype_idx: Optional[ArrayLike] = None,
    ) -> Tuple[NDArray[np.int8], NDArray[np.float32]]: ...

    def multiprocess_read(
        self,
        contig: str,
        start_idxs: ArrayLike,
        end_idxs: ArrayLike,
        sample_idx: Optional[ArrayLike] = None,
        haplotype_idx: Optional[ArrayLike] = None,
        n_jobs: int = 1,
    ) -> NDArray[np.int8]:
        """Read genotypes in parallel from a contig from index i to j, 0-based exclusive.

        Parameters
        ----------
        contig : str
            Name of the contig/chromosome.
        start_idxs : NDArray[np.intp]
            Start indices, 0-based.
        end_idxs : NDArray[np.intp]
            End indices, 0-based exclusive.
        sample_idx : NDArray[np.intp], optional
            Indices of the samples to include. Must be unique.
        haplotype_idx : NDArray[np.intp], optional
            Indices of the haplotypes to include. Must be unique.
        n_jobs : int, optional
            Number of jobs to run in parallel.

        Returns
        -------
        genotypes : NDArray[np.int8]
            Shape: (samples ploidy variants). Genotypes for each query region.
        """
        start_idxs = np.atleast_1d(np.asarray(start_idxs, dtype=np.int32))
        end_idxs = np.atleast_1d(np.asarray(end_idxs, dtype=np.int32))
        n_jobs = min(n_jobs, len(start_idxs))
        starts = np.array_split(start_idxs, n_jobs)
        ends = np.array_split(end_idxs, n_jobs)

        self.close()  # close any open handles so they aren't pickled and cause an error
        tasks = [
            joblib.delayed(self.read)(contig, s, e, sample_idx, haplotype_idx)
            for s, e in zip(starts, ends)
        ]
        with joblib.Parallel(n_jobs=n_jobs) as parallel:
            split_genos = parallel(tasks)
        genos = np.concatenate(split_genos, axis=-1)  # type: ignore
        return genos

    def multiprocess_read_genos_and_dosages(
        self,
        contig: str,
        start_idxs: ArrayLike,
        end_idxs: ArrayLike,
        dosage_field: str,
        sample_idx: Optional[ArrayLike] = None,
        haplotype_idx: Optional[ArrayLike] = None,
        n_jobs: int = 1,
    ) -> Tuple[NDArray[np.int8], NDArray[np.float32]]:
        """Read genotypes in parallel from a contig from index i to j, 0-based exclusive.

        Parameters
        ----------
        contig : str
            Name of the contig/chromosome.
        start_idxs : NDArray[np.intp]
            Start indices, 0-based.
        end_idxs : NDArray[np.intp]
            End indices, 0-based exclusive.
        sample_idx : NDArray[np.intp], optional
            Indices of the samples to include. Must be unique.
        haplotype_idx : NDArray[np.intp], optional
            Indices of the haplotypes to include. Must be unique.
        n_jobs : int, optional
            Number of jobs to run in parallel.

        Returns
        -------
        genotypes : NDArray[np.int8]
            Shape: (samples ploidy variants). Genotypes for each query region.
        """
        start_idxs = np.atleast_1d(np.asarray(start_idxs, dtype=np.int32))
        end_idxs = np.atleast_1d(np.asarray(end_idxs, dtype=np.int32))
        n_jobs = min(n_jobs, len(start_idxs))
        starts = np.array_split(start_idxs, n_jobs)
        ends = np.array_split(end_idxs, n_jobs)

        self.close()  # close any open handles so they aren't pickled and cause an error
        tasks = [
            joblib.delayed(self.read_genos_and_dosages)(
                contig, s, e, dosage_field, sample_idx, haplotype_idx
            )
            for s, e in zip(starts, ends)
        ]
        with joblib.Parallel(n_jobs=n_jobs) as parallel:
            # (s p v), (s v)
            results = parallel(tasks)
        split_genos, split_dosages = zip(*results)
        genos = np.concatenate(split_genos, axis=-1)  # type: ignore
        dosages = np.concatenate(split_dosages, axis=-1)  # type: ignore
        return genos, dosages


class PgenGenos:
    chunked = False
    ploidy = 2
    samples: NDArray[np.str_]

    @property
    def n_samples(self) -> int:
        return len(self.samples)

    def __init__(
        self,
        paths: Union[str, Path, Dict[str, Path]],
        sample_names: NDArray[np.str_],
        contigs: List[str],
    ) -> None:
        if isinstance(paths, (str, Path)):
            paths = {"_all": Path(paths)}

        if len(paths) == 0:
            raise ValueError("No paths provided.")
        n_samples = None
        n_variants = 0
        for p in paths.values():
            if not p.exists():
                raise FileNotFoundError(f"PGEN file {p} not found.")
            if n_samples is None:
                n_samples = pgenlib.PgenReader(bytes(p)).get_raw_sample_ct()  # type: ignore
            n_variants += pgenlib.PgenReader(bytes(p)).get_variant_ct()  # type: ignore
        self.paths = paths
        self.n_variants = n_variants
        self.samples = sample_names
        self.contigs = contigs
        self.handles: Optional[Dict[str, "pgenlib.PgenReader"]] = None
        self.current_sample_idx: Optional[NDArray[np.uint32]] = None

    def _pgen(self, contig: str, sample_idx: Optional[NDArray[np.uint32]]):
        return pgenlib.PgenReader(bytes(self.paths[contig]), sample_subset=sample_idx)  # type: ignore

    def close(self):
        if self.handles is not None:
            for handle in self.handles.values():
                handle.close()
        self.handles = None
        self.current_sample_idx = None

    def read(
        self,
        contig: str,
        start_idxs: ArrayLike,
        end_idxs: ArrayLike,
        sample_idx: Optional[ArrayLike] = None,
        haplotype_idx: Optional[ArrayLike] = None,
    ) -> NDArray[np.int8]:
        changed_sample_idx = False
        if sample_idx is not None:
            _sample_idx = np.atleast_1d(np.asarray(sample_idx, dtype=np.uint32))
            sample_sorter = np.argsort(_sample_idx)
            if self.current_sample_idx is None:
                self.current_sample_idx = _sample_idx[sample_sorter]
                changed_sample_idx = True
            elif np.array_equal(_sample_idx[sample_sorter], self.current_sample_idx):
                sample_sorter = slice(None)
        else:
            self.current_sample_idx = np.arange(self.n_samples, dtype=np.uint32)
            sample_sorter = slice(None)

        if haplotype_idx is not None:
            _haplotype_idx = np.atleast_1d(np.asarray(haplotype_idx, dtype=np.intp))
            if np.array_equal(haplotype_idx, np.arange(self.ploidy)):
                _haplotype_idx = slice(None)
        else:
            _haplotype_idx = slice(None)

        if self.handles is None or changed_sample_idx and "_all" in self.paths:
            handle = self._pgen("_all", self.current_sample_idx)
            self.handles = {c: handle for c in self.contigs}
        elif self.handles is None or changed_sample_idx:
            self.handles = {
                c: self._pgen(c, self.current_sample_idx) for c in self.contigs
            }

        start_idxs = np.atleast_1d(np.asarray(start_idxs, dtype=np.int32))
        end_idxs = np.atleast_1d(np.asarray(end_idxs, dtype=np.int32))

        vars_per_region = end_idxs - start_idxs
        n_vars = vars_per_region.sum()
        rel_start_idxs = _get_rel_starts(start_idxs, end_idxs)
        rel_end_idxs = rel_start_idxs + vars_per_region
        # (v s*2)
        genotypes = np.empty(
            (n_vars, len(self.current_sample_idx) * self.ploidy), dtype=np.int32
        )

        for s, e, rel_s, rel_e in zip(
            start_idxs, end_idxs, rel_start_idxs, rel_end_idxs
        ):
            if s == e:
                continue
            self.handles[contig].read_alleles_range(
                s, e, allele_int32_out=genotypes[rel_s:rel_e]
            )

        # (v s*2) -> (v s 2) -> (s 2 v)
        genotypes = (
            genotypes.astype(np.int8)
            .reshape(n_vars, len(self.current_sample_idx), self.ploidy)
            .transpose(1, 2, 0)
        )

        genotypes = genotypes[sample_sorter, _haplotype_idx]

        return genotypes


class VCFGenos:
    chunked = False
    ploidy = 2
    samples: NDArray[np.str_]

    @property
    def n_samples(self) -> int:
        return len(self.samples)

    def __init__(
        self, vcfs: Union[str, Path, Dict[str, Path]], contig_offsets: Dict[str, int]
    ) -> None:
        if isinstance(vcfs, (str, Path)):
            vcfs = {"_all": Path(vcfs)}

        self.paths = vcfs
        samples = None
        for p in self.paths.values():
            if not p.exists():
                raise FileNotFoundError(f"VCF file {p} not found.")
            if samples is None:
                f = cyvcf2.VCF(str(p))  # type: ignore
                samples = np.array(f.samples)
                f.close()
                del f
        self.samples = samples  # type: ignore
        self.handles: Optional[Dict[str, "cyvcf2.VCF"]] = None
        self.contig_offsets = contig_offsets

    def close(self):
        if self.handles is not None:
            for handle in self.handles.values():
                handle.close()
        self.handles = None

    def read(
        self,
        contig: str,
        starts: ArrayLike,
        ends: ArrayLike,
        sample_idx: Optional[ArrayLike] = None,
        haplotype_idx: Optional[ArrayLike] = None,
        n_variants: NDArray[np.int32] | None = None,
    ) -> NDArray[np.int8]:
        starts = np.atleast_1d(np.asarray(starts, dtype=np.int32))
        ends = np.atleast_1d(np.asarray(ends, dtype=np.int32))

        if self.handles is None and "_all" not in self.paths:
            self.handles = {
                c: cyvcf2.VCF(str(p), lazy=True, threads=len(os.sched_getaffinity(0)))  # type: ignore
                for c, p in self.paths.items()
            }
        elif self.handles is None:
            first_path = next(iter(self.paths.values()))
            handle = cyvcf2.VCF(
                str(first_path), lazy=True, threads=len(os.sched_getaffinity(0))
            )  # type: ignore
            self.handles = {c: handle for c in handle.seqnames}

        if sample_idx is None:
            _sample_idx = slice(None)
        else:
            _sample_idx = np.atleast_1d(np.asarray(sample_idx, dtype=np.int32))

        if haplotype_idx is None:
            _haplotype_idx = slice(None)
        else:
            _haplotype_idx = np.atleast_1d(np.asarray(haplotype_idx, dtype=np.int32))

        if n_variants is None:
            # (r)
            n_variants = np.array(
                [
                    sum(
                        v.is_snp or v.is_indel
                        for v in self.handles[contig](f"{contig}:{s}-{e}")
                    )
                    for s, e in zip(starts + 1, ends)  # vcf is 1-based
                ]
            )

        # (s p v)
        genos = np.empty((self.n_samples, self.ploidy, n_variants.sum()), dtype=np.int8)

        if genos.size == 0:
            return genos

        offsets = _lengths_to_offsets(n_variants)
        for s, e, o in zip(starts + 1, ends, offsets[:-1]):
            for i, v in enumerate(self.handles[contig](f"{contig}:{s}-{e}"), start=o):
                if v.is_snp or v.is_indel:
                    # (s p)
                    genos[..., i] = v.genotype.array()[:, : self.ploidy]

        genos = genos[_sample_idx, _haplotype_idx]
        # cyvcf2 encoding: 0, 1, -1 => gvl/pgen encoding: 0, 1, -9
        genos[genos == -1] = -9

        return genos

    def read_genos_and_ccfs(
        self,
        contig: str,
        starts: ArrayLike,
        ends: ArrayLike,
        ccf_field: str,
        sample_idx: Optional[ArrayLike] = None,
        haplotype_idx: Optional[ArrayLike] = None,
        n_variants: NDArray[np.int32] | None = None,
    ) -> tuple[NDArray[np.int8], NDArray[np.float32]]:
        starts = np.atleast_1d(np.asarray(starts, dtype=np.int32))
        ends = np.atleast_1d(np.asarray(ends, dtype=np.int32))

        if self.handles is None and "_all" not in self.paths:
            self.handles = {
                c: cyvcf2.VCF(str(p), lazy=True, threads=len(os.sched_getaffinity(0)))  # type: ignore
                for c, p in self.paths.items()
            }
        elif self.handles is None:
            first_path = next(iter(self.paths.values()))
            handle = cyvcf2.VCF(
                str(first_path), lazy=True, threads=len(os.sched_getaffinity(0))
            )  # type: ignore
            self.handles = {c: handle for c in handle.seqnames}

        if sample_idx is None:
            _sample_idx = slice(None)
        else:
            _sample_idx = np.atleast_1d(np.asarray(sample_idx, dtype=np.int32))

        if haplotype_idx is None:
            _haplotype_idx = slice(None)
        else:
            _haplotype_idx = np.atleast_1d(np.asarray(haplotype_idx, dtype=np.int32))

        if n_variants is None:
            # (r)
            n_variants = np.array(
                [
                    sum(
                        v.is_snp or v.is_indel
                        for v in self.handles[contig](f"{contig}:{s}-{e}")
                    )
                    for s, e in zip(starts + 1, ends)  # vcf is 1-based
                ]
            )
        # (s p v)
        genos = np.empty((self.n_samples, self.ploidy, n_variants.sum()), dtype=np.int8)
        # (s v)
        ccfs = np.empty((self.n_samples, n_variants.sum()), dtype=np.float32)

        if genos.size == 0:
            return genos, ccfs

        offsets = _lengths_to_offsets(n_variants)
        for s, e, o in zip(starts + 1, ends, offsets[:-1]):
            for i, v in enumerate(self.handles[contig](f"{contig}:{s}-{e}"), start=o):
                if v.is_snp or v.is_indel:
                    # (s p)
                    genos[..., i] = v.genotype.array()[:, : self.ploidy]
                    d = v.format(ccf_field)
                    if d is None:
                        raise CCFFieldError(
                            f"CCF field '{ccf_field}' not found for record {repr(v)}"
                        )
                    # (s, 1, 1) or (s, 1)? -> (s)
                    ccfs[..., i] = d.squeeze()

        genos = genos[_sample_idx, _haplotype_idx]
        # cyvcf2 encoding: 0, 1, -1 => gvl/pgen encoding: 0, 1, -9
        genos[genos == -1] = -9

        ccfs = ccfs[_sample_idx]

        return genos, ccfs

    def init_handles(self) -> Dict[str, "cyvcf2.VCF"]:
        if self.handles is None and "_all" not in self.paths:
            handles = {
                c: cyvcf2.VCF(str(p), lazy=True)  # pyright: ignore
                for c, p in self.paths.items()
            }
        elif self.handles is None:
            handle = cyvcf2.VCF(str(next(iter(self.paths.values()))), lazy=True)  # pyright: ignore
            handles = {c: handle for c in handle.seqnames}
        else:
            handles = self.handles
        return handles

    def __del__(self) -> None:
        if self.handles is not None:
            for handle in self.handles.values():
                handle.close()



import json
from enum import Enum
from pathlib import Path

import typer


class Arch(str, Enum):
    CPU = "cpu"
    GPU = "gpu"


def main(
    n_queries: int,
    min_intervals: int,
    max_intervals: int,
    min_length: int,
    max_length: int,
    min_width: int,
    max_width: int,
    max_value: float,
    ti_arch: Arch,
    out: Path,
):
    from time import perf_counter

    import numpy as np
    import polars as pl
    import taichi as ti
    from genvarloader.dataset.intervals import (
        intervals_to_tracks,
        ti_intervals_to_tracks,
    )
    from genvarloader.types import INTERVAL_DTYPE
    from genvarloader.utils import lengths_to_offsets
    from loguru import logger
    from tqdm.auto import tqdm
    from utils import nonoverlapping_intervals

    ti.init(ti.gpu if ti_arch is Arch.GPU else ti.cpu)

    rng = np.random.default_rng(0)

    regions = np.empty((n_queries, 4), dtype=np.int32)
    n_itvs_per_query = rng.integers(min_intervals, max_intervals, size=n_queries)
    intervals = np.empty(n_itvs_per_query.sum(), dtype=INTERVAL_DTYPE)
    offsets = lengths_to_offsets(n_itvs_per_query)
    offset_idxs = np.arange(len(offsets) - 1, dtype=np.int32)

    logger.info("Generating intervals")
    for query in range(n_queries):
        n_intervals = n_itvs_per_query[query]
        offset = offsets[query]
        length = rng.integers(min_length, max_length)
        seed = rng.integers(0, 2**31)
        regions[query] = np.array([0, 0, length, 1], dtype=np.int32)
        coordinates = nonoverlapping_intervals(
            n_intervals, 0, length, min_width, max_width, seed
        )
        values = np.random.rand(n_intervals).astype(np.float32) * max_value
        intervals[offset : offset + n_intervals]["start"] = coordinates[:, 0]
        intervals[offset : offset + n_intervals]["end"] = coordinates[:, 1]
        intervals[offset : offset + n_intervals]["value"] = values

    logger.info("Calling once for compilation")
    _ = intervals_to_tracks(
        offset_idxs=offset_idxs,
        regions=regions,
        intervals=intervals,
        offsets=offsets,
    )

    _ = ti_intervals_to_tracks(
        offset_idxs=offset_idxs,
        regions=regions,
        intervals=intervals,
        offsets=offsets,
    )

    n_loops = 50
    nb_times = np.empty(n_loops, dtype=np.float64)
    ti_times = np.empty(n_loops, dtype=np.float64)
    for i in tqdm(range(n_loops), desc="Benchmarking"):
        t0 = perf_counter()
        nb_tracks = intervals_to_tracks(
            offset_idxs=offset_idxs,
            regions=regions,
            intervals=intervals,
            offsets=offsets,
        )
        nb_times[i] = perf_counter() - t0

        t0 = perf_counter()
        ti_tracks = ti_intervals_to_tracks(
            offset_idxs=offset_idxs,
            regions=regions,
            intervals=intervals,
            offsets=offsets,
        )
        ti_times[i] = perf_counter() - t0

        np.testing.assert_equal(nb_tracks, ti_tracks)
    pl.DataFrame(
        {
            "numba": nb_times,
            f"taichi_{ti_arch}": ti_times,
        }
    ).write_csv(out.with_suffix(".csv"))
    with open(out.with_suffix(".json"), "w") as f:
        metadata = {
            "n_queries": n_queries,
            "min_intervals": min_intervals,
            "max_intervals": max_intervals,
            "min_length": min_length,
            "max_length": max_length,
            "min_width": min_width,
            "max_width": max_width,
            "max_value": max_value,
            "ti_arch": ti_arch,
        }
        json.dump(metadata, f, indent=4)


if __name__ == "__main__":
    typer.run(main)
